import scipy as sp
import scipy.linalg as la
import scipy.stats as stats
import numpy as np
import pandas as pd
from distutils.util import strtobool
from math import sqrt


def boolanize(string):
    r"""
    Convert command line parameter "True"/"False" into boolean

    Arguments:
        string (string):
        "False" or "True"

    Returns:
        (bool):

            False/True
    """
    return bool(strtobool(string))


def nans(shape):
    r"""
    Create numpy array of NaNs

    Arguments:
        shape (tuple):
            shape of the empty array

    Returns:
        (numpy array):
            numpy array of NaNs
    """
    a = np.empty(shape, dtype=float)
    a.fill(np.nan)
    return a


def scale(x):
    r"""
    Mean center and unit variance input array

    Arguments:
        x (array-like):
            array to be scaled by column
    Returns:
        (numpy array):
            mean-centered, unit-variance array of x
    """

    x = np.array(x)
    x -= x.mean(axis=0)
    x /= x.std(axis=0)
    return x


def getEigen(covMatrix, reverse=True):
    r""" Get eigenvectors and values of hermitian matrix:

    Arguments:
        covMatrix (array-like):
            hermitian matrix
        reverse (bool):
            if True (default): order eigenvalues (and vectors) in
            decreasing order

    Returns:
        (tuple):
            tuple containing:

            - eigenvectors
            - eigenvalues
    """
    covMatrix = np.array(covMatrix)

    S, U = la.eigh(covMatrix)
    if reverse:
        S = S[::-1]
        U = U[:, ::-1]
    return (S, U)


def getVariance(eigenvalue):
    r"""
    Based on input eigenvalue computes cumulative sum and normalizes to overall
    sum to obtain variance explained

    Arguments:
        eigenvalue (array-like):
            eigenvalues
    Returns:
        (float):
            variance explained
    """

    v = np.array(eigenvalue).cumsum()
    v /= v.max()
    return (v)


def regularize(m, verbose=True):
    r"""
    Make matrix positive-semi definite by ensuring minimum eigenvalue >= 0:
    add absolute value of minimum eigenvalue and 1e-4 (for numerical stability
    of abs(min(eigenvalue) < 1e-4 to diagonal of matrix

    Arguments:
        m (array-like):
            symmetric matrix

    Returns:
       (tuple):
            Returns tuple containing:

            - positive, semi-definite matrix from input m (numpy array)
            - minimum eigenvalue of input m
    """
    S, U = la.eigh(m)
    minS = S.min()
    if minS < 0:
        verboseprint(
            "Regularizing: minimum Eigenvalue %6.4f" % S.min(),
            verbose=verbose)
        m += (abs(S.min()) + 1e-4) * sp.eye(m.shape[0])
    elif minS < 1e-4:
        verboseprint("Make numerically stable: minimum Eigenvalue %6.4f" %
                     S.min(), verbose=verbose)
        m += 1e-4 * sp.eye(m.shape[0])
    else:
        verboseprint("Minimum Eigenvalue %6.4f" % S.min(), verbose=verbose)

    return (m, minS)


def generate_permutation(P, S, n, seed=12321, exclude_zero=False):
    r"""
    Generate permutation.

    Arguments:
        seed (int):
            used as seed for pseudo-random numbers generation; default: 12321
        n (int):
            number of permutations to generated
        P (int):
            total number of traits
        S (int):
            subsampling size
        exclude_zero (bool):
            should zero be in set to draw from

    Returns:
        (list):
            Returns list of length n containing [np.arrays] of length [`S`]
            with subsets/permutations of numbers range(P)
    """
    rand_state = np.random.RandomState(seed)
    return_list = [None] * n
    if exclude_zero:
        rangeP = range(P)[1:]
    else:
        rangeP = range(P)
    for i in xrange(n):
        perm_dic = rand_state.choice(a=rangeP, size=S, replace=False)
        return_list[i] = perm_dic
    return return_list


def inflate_matrix(bootstrap_traits, bootstrap, P, zeros=True):
    r"""
    Project small matrix into large matrix using indeces provided:

    Arguments:
        bootstrap_traits (array-like):
            [`S` x `S`] covariance matrix estimates
        bootstrap (array-like):
            [`S` x 1] array with indices to project [`S` x `S`] matrix values
            into [`P` x `P`] matrix
        P (int):
            total number of dimensions
        zeros (bool):
            fill void spaces in large matrix with zeros (True,
            default) or nans (False)

    Returns:
        (numpy array):
            Returns [`P` x `P`] matrix containing [`S` x `S`] matrix values at
            bootstrap indeces and zeros/nans elswhere
    """
    index = np.ix_(np.array(bootstrap), np.array(bootstrap))
    if zeros is True:
        all_traits = np.zeros((P, P))
    else:
        all_traits = nans((P, P))
    all_traits[index] = bootstrap_traits
    return (all_traits)


def verboseprint(message, verbose=True):
    r"""
    Print message if verbose option is True.

    Arguments:
        message (string):
            text to print
        verbose (bool):
            flag whether to print message (True) or not (False)
    """
    if verbose is True:
        print message


def match(samples_ref, data_compare, samples_compare, squarematrix=False):
    r"""
    Match the order of data and ID matrices to a reference sample order,

    Arguments:
        samples_ref (array-like):
            [`M`] sammple Ids used as reference
        data_compare (array-like):
            [`N` x `L`] data matrix with [`N`] samples and [`L`] columns
        samples_compare (array-like):
            [`N`] sample IDs to be matched to samples_ref
        squarematrix (bool):
            is data_compare a square matrix i.e. samples in cols and rows

    Returns:
        (tuple):
            tuple containing:

            - data_compare (numpy array):
              [`M` x `L`] data matrix of input data_compare
            - samples_compare (numpy array):
              [`M`] sample IDs of input samples_compare
            - samples_before (int):
              number of samples in data_compare/samples_compare before matching
              to samples_ref
            - samples_after (int):
              number of samples in data_compare/samples_compare after matching
              to samples_ref
    """
    samples_before = samples_compare.shape[0]
    subset = pd.match(samples_ref, samples_compare)

    data_compare = data_compare[subset, :]
    if squarematrix:
        data_compare = data_compare[:, subset]
    samples_compare = samples_compare[subset]
    samples_after = samples_compare.shape[0]
    np.testing.assert_array_equal(
        samples_ref,
        samples_compare,
        err_msg=("Col order does not match. These"
                 "are the differing columns:\t%s") %
        (np.array_str(np.setdiff1d(samples_ref, samples_compare))))
    return (data_compare, samples_compare, samples_before, samples_after)


def AlleleFrequencies(snp):
    hc_snps = np.array([makeHardCalledGenotypes(s) for s in snp])
    counts = np.array(np.unique(hc_snps, return_counts=True))
    frequencies = counts[1, :] / float(len(hc_snps))
    major_a = sqrt(frequencies.max())
    minor_a = 1 - major_a
    return minor_a, major_a


def makeHardCalledGenotypes(snp):
    if snp <= 0.5:
        return 0
    elif snp > 1.5:
        return 2
    else:
        return 1


def effectiveTests(test):
    # 1. get correlation matrix
    corr_matrix = stats.spearmanr(test).correlation
    # 2. Get eigenvalues of correlation matrix:
    eigenval, eigenvec = la.eigh(corr_matrix)
    # 3. Determine effective number of tests:
    t = np.sqrt(eigenval).sum()**2 / eigenval.sum()
    return t



import sys

import scipy as sp
import pandas as pd
import numpy as np

import limix.qtl as qtl
import limix.plot as plot
import limmbo.plot.manhattan as mh
import limix.mtset

from limmbo.utils.utils import verboseprint
from limmbo.utils.utils import effectiveTests

import matplotlib as mpl
mpl.use('Agg')


class GWAS(object):
    """
    """

    def __init__(self, datainput, seed=10, verbose=True,
                 searchDelta=False):
        '''
        nothing to initialize
        '''
        self.verbose = verbose
        self.seed = seed
        self.searchDelta = searchDelta
        self.genotypes = datainput.genotypes
        self.genotypes_info = datainput.genotypes_info
        self.phenotypes = datainput.phenotypes
        self.phenotype_ID = datainput.phenotype_ID
        self.covariates = datainput.covariates
        self.relatedness = datainput.relatedness
        self.pcs = datainput.pcs
        self.Cg = datainput.Cg
        self.Cn = datainput.Cn
        self.test = "lrt"

        self.pvalues = None
        self.pvalues_adjust = None
        self.pvalues_empirical_raw = None
        self.stats = None
        self.z = None
        self.model = None
        self.pvalues_empirical = None
        self.adjustBy = None
        self.estimate_vd = None
        self.fdr_empirical = None

        try:
            self.genotypes = np.array(self.genotypes)
        except:
            raise IOError("datainput.genotypes cannot be coverted to np.array")

        try:
            self.phenotypes = np.array(self.phenotypes)
        except:
            raise IOError(
                "datainput.phenotypes cannot be coverted to np.array")

        if self.covariates is not None:
            try:
                self.covariates = np.array(self.covariates)
            except:
                raise IOError(
                    "datainput.covariates cannot be coverted to np.array")

        if self.relatedness is not None:
            try:
                self.relatedness = np.array(self.relatedness)
            except:
                raise IOError(
                    "datainput.relatedness cannot be coverted to np.array")

    def runAssociationAnalysis(self, mode, setup="lmm",
                               adjustSingleTrait=None):
        r"""
        Analysing the association between phenotypes, genotypes, optional
        covariates and random genetic effects.

        Arguments:
            mode (string):
                pecifies the type of linear model: either 'multitrait' for
                multivariate analysis or 'singletrait' for univariate analysis.
            setup (string):
                specifies the linear model setup: either 'lmm' for linear mixed
                model or 'lm' for a simple linear model.
            adjustSingleTrait (string):
                Method to adjust single-trait association p-values for testing
                multiple traits; If None (default) no adjusting. Options are
                'bonferroni' (for bonferroni correction') or 'effective' (for
                correcting for the effective number of tests as described in
                `(Galwey,2009) <http://onlinelibrary.wiley.com/doi/10.
                1002/gepi.20408/abstract>`_.


        Returns:
            (dictionary):
                dictionary containing:

                - **lm** (:class:`limix.qtl.LMM`):
                  LIMIX LMM object
                - **pvalues** (numpy array):
                  [`NrSNP` x `P`] (when mode is singletrait) or [1 x`NrSNP`]
                  array of p-values.
                - **betas** (numpy array):
                  [`NrSNP` x `P`] array of effect size estimates per SNP across
                  all traits.
                - **pvalues_adjust** (numpy array):
                  only returned if mode is 'singletrait' and
                  'adjustSingleTrait' is not None; contains single-trait
                  p-values adjusted for the number of single-trait analyses
                  conducted.

        Examples:

            .. doctest::

                >>> from pkg_resources import resource_filename
                >>> from limmbo.io.reader import ReadData
                >>> from limmbo.io.input import InputData
                >>> from limmbo.core.gwas import GWAS
                >>> data = ReadData(verbose=False)
                >>> file_pheno = resource_filename('limmbo',
                ...                                'io/test/data/pheno.csv')
                >>> file_geno = resource_filename(
                ...     'limmbo', 'io/test/data/genotypes.csv')
                >>> file_relatedness = resource_filename('limmbo',
                ...                     'io/test/data/relatedness.csv')
                >>> file_covs = resource_filename('limmbo',
                ...                               'io/test/data/covs.csv')
                >>> file_Cg = resource_filename('limmbo',
                ...                     'io/test/data/Cg.csv')
                >>> file_Cn = resource_filename('limmbo',
                ...                     'io/test/data/Cn.csv')
                >>> data.getPhenotypes(file_pheno=file_pheno)
                >>> data.getCovariates(file_covariates=file_covs)
                >>> data.getRelatedness(file_relatedness=file_relatedness)
                >>> data.getGenotypes(file_genotypes=file_geno)
                >>> data.getVarianceComponents(file_Cg=file_Cg,
                ...                            file_Cn=file_Cn)
                >>> indata = InputData(verbose=False)
                >>> indata.addPhenotypes(phenotypes = data.phenotypes)
                >>> indata.addRelatedness(relatedness = data.relatedness)
                >>> indata.addCovariates(covariates = data.covariates)
                >>> indata.addGenotypes(genotypes=data.genotypes,
                ...                     genotypes_info=data.genotypes_info)
                >>> indata.addVarianceComponents(Cg = data.Cg, Cn=data.Cn)
                >>> indata.commonSamples()
                >>> indata.regress()
                >>> indata.transform(transform="scale")
                >>> gwas = GWAS(datainput=indata, seed=10, verbose=False)
                >>>
                >>>
                >>> # Example of multi-trait single-variant association testing
                >>> # using a linear mixed model.
                >>> resultsAssociation = gwas.runAssociationAnalysis(
                ...     setup="lmm", mode="multitrait")
                >>> resultsAssociation.keys()
                ['lm', 'betas', 'pvalues']
                >>> resultsAssociation['pvalues'].shape
                (1, 20)
                >>> '{:0.3e}'.format(resultsAssociation['pvalues'].min())
                '4.584e-09'
                >>> resultsAssociation['betas'].shape
                (10, 20)
                >>>
                >>>
                >>> # Example of single-trait single-variant association
                >>> # testing using a linear mixed model.
                >>> resultsAssociation = gwas.runAssociationAnalysis(
                ...     setup="lmm", mode="singletrait",
                ...     adjustSingleTrait = "effective")
                >>> resultsAssociation.keys()
                ['pvalues_adjust', 'lm', 'betas', 'pvalues']
                >>> resultsAssociation['pvalues'].shape
                (10, 20)
                >>> resultsAssociation['betas'].shape
                (10, 20)
                >>> '{:0.3e}'.format(
                ...     resultsAssociation['pvalues_adjust'].min())
                '2.262e-03'
        """

        # set parameters for the analysis
        self.N, self.P = self.phenotypes.shape
        self.S = self.genotypes.shape[1]
        verboseprint("Loaded {} samples, {} phenotypes, {} snps".format(
            self.N, self.P, self.S), verbose=self.verbose)

        self.setup = setup
        self.mode = mode

        if mode == "multitrait":
            associationResults = self.__multiTraitAssociation_anyeffect(
                genotypes=self.genotypes)

        if mode == "singletrait":
            associationResults = self.__singleTraitAssociation(
                genotypes=self.genotypes,
                adjustSingleTrait=adjustSingleTrait)

        return associationResults

    def __multiTraitAssociation_anyeffect(self, genotypes, empiricalP=False,
                                          computeFDR=False):
        r"""
        Muti-trait association test, testing for an effect of the genotype on
        `any` phenotype (`any` effect test). Wraps around
        `qtl.qtl_test_lmm_kronecker`.

        Arguments:
            genotypes (array-like):
                [`N` x `NrSNP`] genotype array of [`N`] samples and [`NrSNP`]
                genotypes
            empiricalP (bool):
                set to True if association test is part of estimating empirical
                pvalues
            computeFDR (bool):
                set to True if association test is part of estimating empirical
                FDR

        Returns:
            (dictionary):
                dictionary containing:

                - **lm**(:class:`limix.qtl.LMM`):
                  LIMIX LMM object
                - **pvalues** (numpy array):
                  [1 x`NrSNP`] array of
                  p-values
                - **betas** (numpy array):
                  [`NrSNP` x `P`] array of effect size estimates per SNP across
                  all traits

        """
        if self.covariates is None:
            Acovs = None
        else:
            Acovs = sp.eye(self.P)
        if self.setup is "lmm":
            if self.Cg is None:
                if self.P > 30:
                    print("Warning: For large trait sizes, computation times "
                          "for pure REML variance decomposition are long, "
                          "consider bootstrapping trait-trait covariance"
                          " components")
                verboseprint("Estimate Variance components",
                             verbose=self.verbose)
                self.__varianceDecomposition()
            K1c = self.Cg
            K2c = self.Cn
            K1r = self.relatedness
            self.model = "lmm_mt"
        else:
            K1c = 1e-9 * sp.eye(self.P)
            K2c = sp.cov(self.phenotypes.T)
            K1r = sp.eye(self.N)

            if self.pcs is not None:
                self.model = "lm_mt_pcs"
            else:
                self.model = "lm_mt"
        Asnps = sp.eye(self.P)
        if not empiricalP and not computeFDR:
            verboseprint("Computing multi-trait (any effect) model: {}".format(
                self.model), verbose=self.verbose)

        lm, pvalues = qtl.qtl_test_lmm_kronecker(snps=genotypes,
                                                 phenos=self.phenotypes,
                                                 Asnps=Asnps, Acovs=Acovs,
                                                 covs=self.covariates, K1r=K1r,
                                                 K1c=K1c, K2c=K2c,
                                                 searchDelta=self.searchDelta)

        if not empiricalP and not computeFDR:
            betas = lm.getBetaSNP()
        else:
            betas = None
            lm = None

        return {"pvalues": pvalues, "betas": betas, "lm": lm}

    def __singleTraitAssociation(self, genotypes, adjustSingleTrait=None,
                                 empiricalP=False, computeFDR=False):
        r"""
        Single-trait association test. Wraps around `qtl.qtl_test_lmm`.

        Arguments:
            genotypes (array-like):
                [`N` x `NrSNP`] genotype array of [`N`] samples and [`NrSNP`]
                genotypes
            empiricalP (bool):
                set to True if association test is part of estimating empirical
                pvalues
            computeFDR (bool):
                set to True if association test is part of estimating empirical
                FDR
            adjustSingleTrait (string):
                Method to adjust single-trait association p-values for testing
                multiple traits; If None (default) no adjusting. Options are
                'bonferroni' (for bonferroni correction') or 'effective' (for
                correcting for the effective number of tests as described in
                `(Galwey,2009) <http://onlinelibrary.wiley.com/doi/10.
                1002/gepi.20408/abstract>`_.

        Returns:
            (dictionary):
                dictionary containing:

                    - **lm**(:class:`limix.qtl.LMM`):
                      LIMIX LMM object
                    - **pvalues** (numpy array):
                      [`P` x `NrSNP`] array of p-values
                    - **betas** (numpy array):
                      [`P` x `NrSNP`] array of effect size estimates per SNP
                      across all traits
                    - **pvalues_adjust** (numpy array):
                      only returned if 'adjustSingleTrait' is not None;
                      contains single-trait p-values adjusted for the number of
                      single-trait analyses conducted.

        """

        if self.setup is "lmm":
            self.model = "lmm_st"
            K = self.relatedness
        else:
            if self.pcs is not None:
                self.model = "lm_st_pcs"
            else:
                self.model = "lm_st"
                K = None

        if not empiricalP and not computeFDR:
            verboseprint("Computing single-trait association ({})".format(
                self.model), verbose=self.verbose)
            lm = qtl.qtl_test_lmm(snps=genotypes, pheno=self.phenotypes,
                                  K=K, covs=self.covariates, test=self.test)
            pvalues = lm.getPv()

        if not empiricalP and not computeFDR:
            betas = lm.getBetaSNP()
        else:
            betas = None
            lm = None

        if adjustSingleTrait is not None:
            if self.adjustBy is None:
                if adjustSingleTrait is "bonferroni":
                    self.adjustBy = self.P
                elif adjustSingleTrait is "effective":
                    self.adjustBy = effectiveTests(self.phenotypes)
                else:
                    raise (("{} is not a provided method to adjust "
                            "single-trait pvalues for multiple hypothesis "
                            "testing").format(
                        adjustSingleTrait))
            pvalues_adjust = np.array([self.__adjust(p) for p in pvalues])
        else:
            pvalues_adjust = None

        self.adjustSingleTrait = adjustSingleTrait

        return {"lm": lm, "pvalues": pvalues, "pvalues_adjust": pvalues_adjust,
                "betas": betas}

    def saveAssociationResults(self, results, outdir, name="",
                               pvalues_empirical=None):
        r"""
        Saves results of association analyses.

        Arguments:
            results (dictionary):
                dictionary generated via runAssociation analysis, containing:

                - **lm** (:class:`limix.qtl.LMM`):
                  LIMIX LMM object
                - **pvalues** (numpy array):
                  [`P` x `NrSNP`] array of p-values
                - **betas** (numpy array):
                  [`P` x `NrSNP`] array of effect size estimates per SNP
                  across all traits
                - **pvalues_adjust** (numpy array):
                  only returned mode == singletrait and if
                  'self.adjustSingleTrait' is not None; contains single-trait
                  p-values adjusted for the number of single-trait analyses
                  conducted.

            outdir (string):
                '/path/to/output/directory'; needs user writing permission
            name (string):
                input name specific name, such as chromosome used in analysis
            pvalues_empirical (pandas dataframe, optional):
                empirical pvalues via adjustSingleTrait.

        Returns:
            (None)
        """
        if name is not "":
            self.name = name
            outstring = (outdir, self.model, "_" + name)
        else:
            self.name = ""
            outstring = (outdir, self.model, "")

        self.genotypes_info['SNP'] = self.genotypes_info.index

        beta_df = pd.DataFrame(results['betas'].T,
                               index=self.genotypes_info.index,
                               columns=self.phenotype_ID)
        beta_df = pd.concat([self.genotypes_info, beta_df], axis=1)

        if self.mode is 'singletrait':
            pvalues_df = pd.DataFrame(results['pvalues'].T,
                                      index=self.genotypes_info.index,
                                      columns=self.phenotype_ID)
            pvalues_df = pd.concat([self.genotypes_info, pvalues_df], axis=1)

            if results['pvalues_adjust'] is not None:
                pvalues_adjust_df = pd.DataFrame(
                    results['pvalues_adjust'].T,
                    index=self.genotypes_info.index,
                    columns=self.phenotype_ID)
                pvalues_adjust_df = pd.concat([self.genotypes_info,
                                               pvalues_adjust_df], axis=1)

                pvalues_adjust_df.to_csv("%s/%s_padjust%s.csv" % outstring,
                                         index=False)

            if pvalues_empirical is not None:

                pempirical_df = pd.DataFrame(pvalues_empirical.T,
                                             index=self.genotypes_info.index,
                                             columns=self.phenotype_ID)
                # index=SNP, columns=['Pempirical'])
                pempirical_df = pd.concat([self.genotypes_info,
                                           pempirical_df], axis=1)

                pempirical_raw_df = pd.DataFrame(
                    self.pvalues_empirical_raw.T,
                    index=self.genotypes_info.index,
                    columns=self.phenotype_ID)
                pempirical_raw_df = pd.concat([self.genotypes_info,
                                               pempirical_raw_df], axis=1)

                pempirical_raw_df.to_csv("%s/%s_pempirical_raw%s%s.csv" %
                                         (outstring + (self.fdr,)),
                                         index=False)

        if self.mode is 'multitrait':
            pvalues_df = pd.DataFrame(results['pvalues'].T,
                                      index=self.genotypes_info.index,
                                      columns=["P"])
            pvalues_df = pd.concat([self.genotypes_info, pvalues_df], axis=1)

            if pvalues_empirical is not None:
                pvalues_empirical_df = pd.DataFrame(
                    pvalues_empirical.T, index=self.genotypes_info.index,
                    columns=["P"])
                pvalues_empirical_df = pd.concat(
                    [self.genotypes_info, pvalues_empirical_df], axis=1)

        pvalues_df.to_csv("%s/%s_pvalue%s.csv" % outstring, index=False)
        beta_df.to_csv("%s/%s_betavalue%s.csv" % outstring, index=False)

        if pvalues_empirical is not None:
            pvalues_empirical_df.to_csv("%s/%s_pempirical%s%s.csv" %
                                        (outstring + (self.fdr,)), index=False)

        if self.estimate_vd:
            if self.timeVD is not None:
                pd.DataFrame(self.timeVD).to_csv(
                    "%s/timeVarianceDecomposition_REML.csv" % outdir,
                    index=False, header=False)

            pd.DataFrame(self.Cg).to_csv("%s/Cg_REML.csv" % (outdir),
                                         index=False, header=False)
            pd.DataFrame(self.Cn).to_csv("%s/Cn_REML.csv" % (outdir),
                                         index=False, header=False)

            if self.fdr is not None:
                pd.DataFrame(self.allppermute).to_csv("%s/%s_ppermute%s.csv" %
                                                      outstring, index=False)
                pd.DataFrame(['FDR', str(self.FDR)]).T.to_csv(
                    "%s/%s_empiricalFDR_%s.csv" % outstring, header=False,
                    index=False)

    def computeEmpiricalP(self, pvalues, nrpermutations=1000):
        r"""
        Compute empirical p-values: permute the genotypes, do the
        association test, record if permuted p-value of SNP is smaller than
        original p-value. Sum these occurrences and divide by total number of
        permutation.

        Arguments:
            pvalues (array-like):
                [`P` x `NrSNP`] (single-trait) or [1 x `NrSNP`] (multi-trait)
                array of p-values.
            nrpermutations (int):
                number of permutations; 1/nrpermutations is the maximum level
                of significance (alpha)to test for, e.g.
                nrpermuations=100 -> alpha=0.01

        Returns:
            (numpy array):
                [`P` x `NrSNP`] (single-trait) or [1 x `NrSNP`] (multi-trait)
                array of empirical p-values
        """

        verboseprint("Computing empirical p-values", verbose=self.verbose)

        np.random.seed(seed=self.seed)

        self.nrpermutations = nrpermutations
        pvalues = np.array(pvalues)

        if self.mode is "multitrait":
            count = np.zeros((nrpermutations, self.S))
        else:
            count = np.zeros((nrpermutations, self.P, self.S))

        for ps in range(nrpermutations):
            verboseprint("Permutation %s" % ps)
            genotypes_permute = self.genotypes[np.random.choice(
                self.N, self.N, replace=False), :]
            if self.mode is "multitrait":
                resultsPermute = self.__multiTraitAssociation_anyeffect(
                    genotypes=genotypes_permute, empiricalP=True)
                count[ps, :] = resultsPermute['pvalues'] <= pvalues
            if self.mode is "singletrait":
                resultsPermute = self.__singletraitAssociation(
                    genotypes=genotypes_permute, empiricalP=True)
                count[ps, :, :] = resultsPermute['pvalues'] <= pvalues

        pvalues_empirical = np.sum(count, axis=0) / nrpermutations

        return pvalues_empirical

    def computeFDR(self, fdr):
        r"""
        Create an empirical p-values distribution by permuting the genotypes
        and running the association test on these (10 random permuations).
        Recored the observed pvalues, order by rank and find the rank of the
        desired FDR to determine the empirical FDR.

        Arguments:
            fdr (float):
                desired fdr threshold

        Returns:
            (tuple):
                tuple containing:

                    - **empirical fdr** (float):
                    - **empirical_pvalue_distribution** (numpy array):
                      array of empirical p-values for 10 permutations tests
        """

        verboseprint("Computing empirical FDR", verbose=self.verbose)

        np.random.seed(seed=self.seed)
        # tests = int(1/self.options.fdr)
        tests = 10

        ppermute = np.zeros((tests, self.S))

        for ps in range(tests):
            genotypes_permute = self.genotypes[np.random.choice(
                self.N, self.N, replace=False), :]
            if self.mode == "multitrait":
                resultsFDR = self.__multiTraitAssociation_anyeffect(
                    genotypes=genotypes_permute, computeFDR=True)
                ppermute[ps, :] = resultsFDR['pvalues_permute']
            if self.mode == "singletrait":
                resultsFDR = self.__singleTraitAssociation(
                    genotypes=genotypes_permute, computeFDR=True,
                    adjustSingleTrait=self.adjustSingleTrait)
                pvalues_adjust_min = resultsFDR['pvalues_permute_adjust'].min(
                    axis=0)
                ppermute[ps, :] = pvalues_adjust_min

        SNPsPassingFDR = int(fdr * self.S * tests)
        self.allppermute = ppermute.flatten()
        self.allppermute.sort()
        self.fdr_empirical = self.allppermute[SNPsPassingFDR]

        return self.fdr_empirical, self.allppermute

    def __adjust(self, pv):
        r"""
        Adjust pvalues for multiple hypothesis testing by multiplying with
        a constant (if adjustBy=P, equivalent to Bonferroni adjustment).

        Arguments:
            pv (array like):
                p-values

        Returns:
            pv (numpy array):
                adjusted p-values
        """
        pvadjust = np.array([min(pveff, 1) for pveff in (pv * self.adjustBy)])
        return pvadjust

    def __varianceDecomposition(self, cache=True):
        vd = limix.mtset.MTSet(Y=self.phenotypes, R=self.relatedness)
        vd_null_info = vd.fitNull(n_times=1000, rewrite=True)
        if vd_null_info['conv']:
            verboseprint("Variance decomposition converged")
            self.Cg = vd_null_info['Cg'] + 1e-4 * sp.eye(self.P)
            self.Cn = vd_null_info['Cn'] + 1e-4 * sp.eye(self.P)
            self.time = vd_null_info['time']
        else:
            sys.exit("Variance decomposition did not converge")

    def manhattanQQ(self, results, colourS='DarkBLue', colourNS='Orange',
                    alphaS=1, alphaNS=0.1, thr_plotting=None, saveTo=None):
        r"""
        Plot manhattan and quantile-quantile plot of association results.

        Arguments:
            results (dictionary):
                dictionary generated via runAssociation analysis
            colourS (string, optional):
                colour of significant points
            colourNS (string, optional):
                colour of non-significant points
            alphaS (float, optional):
                plotting transparency of significant points
            alphaNS (float, optional):
                plotting transparency of non-significant points
            thr_plotting (float, optional):
                y-intercept for horizontal line as a marker for significance
            saveTo (string, optional):
                /path/to/output/directory to automatically save plot as pdf;
                needs user writing permission.

        Returns:
            (None)
        """
        import matplotlib.pyplot as plt
        if self.mode is 'singletrait':
            fig = plt.figure(figsize=[12, 4])
            pv_min = np.array(results['pvalues']).min(axis=0)
            data = pd.DataFrame({'pv': pv_min,
                                 'chrom': np.array(self.genotypes_info['chrom']
                                                   ).astype('int'),
                                 'pos': np.array(self.genotypes_info['pos']
                                                 ).astype('int')})
            ax1 = fig.add_subplot(2, 1, 1)
            mh.plot_manhattan(df=data,
                              null_style={'alpha': float(
                                  alphaNS), 'color': colourNS},
                              alt_style={'alpha': float(
                                  alphaS), 'color': colourS},
                              alpha=thr_plotting)
            ax1.set_title('%s' % self.name)
            ax2 = fig.add_subplot(2, 1, 2)
            plot.qqplot(results['pvalues'].ravel(), alphaLevel=None)
            fig.tight_layout()

            if results['pvalues_adjust'] is not None:
                fig = plt.figure(figsize=[12, 8])
                pv_adjust_min = np.array(results['pvalues_adjust']).min(
                    axis=0).ravel()
                data_adjust = pd.DataFrame({'pv': pv_adjust_min,
                                            'chrom': np.array(
                                                self.genotypes_info['chrom']
                                            ).astype('int'),
                                            'pos': np.array(
                                                self.genotypes_info['pos']
                                            ).astype('int')})
                ax1 = fig.add_subplot(4, 1, 1)
                mh.plot_manhattan(df=data,
                                  null_style={'alpha': float(
                                      alphaNS), 'color': colourNS},
                                  alt_style={'alpha': float(
                                      alphaS), 'color': colourS},
                                  alpha=thr_plotting)
                ax1.set_title('%s (p-values)' % self.name)
                ax2 = fig.add_subplot(4, 1, 2)
                plot.qqplot(results['pvalues'].ravel())
                ax1 = fig.add_subplot(4, 1, 3)
                mh.plot_manhattan(df=data_adjust,
                                  null_style={'alpha': float(
                                      alphaNS), 'color': colourNS},
                                  alt_style={'alpha': float(
                                      alphaS), 'color': colourS},
                                  alpha=thr_plotting)
                ax1.set_title('%s (p-values adjust)' % self.name)
                ax2 = fig.add_subplot(4, 1, 4)
                plot.qqplot(results['pvalues_adjust'].ravel())
                fig.tight_layout()

        if self.mode is 'multitrait':
            pv = np.array(results['pvalues']).ravel()
            data = pd.DataFrame({'pv': pv,
                                 'chrom': np.array(self.genotypes_info['chrom']
                                                   ).astype('int'),
                                 'pos': np.array(self.genotypes_info['pos']
                                                 ).astype('int')})
            fig = plt.figure(figsize=[12, 4])
            ax1 = fig.add_subplot(2, 1, 1)
            mh.plot_manhattan(df=data,
                              null_style={'alpha': float(
                                  alphaNS), 'color': colourNS},
                              alt_style={'alpha': float(
                                  alphaS), 'color': colourS},
                              alpha=thr_plotting)
            ax1.set_title('%s' % self.name)
            ax2 = fig.add_subplot(2, 1, 2)
            plot.qqplot(results['pvalues'].ravel())
            fig.tight_layout()

        if saveTo is not None:
            if self.name is not "":
                output = '{}/{}_{}.png'.format(saveTo, self.name, self.model)
            else:
                output = '{}/{}.png'.format(saveTo, self.model)
            fig.savefig(output)



from limmbo.io.parser import getGWASargs
from limmbo.io.reader import ReadData
from limmbo.io.input import InputData
from limmbo.core.gwas import GWAS


def entry_point():

    # parse command-line arguments
    parser = getGWASargs()
    options = parser.parse_args()
    if options.lmm and options.file_relatedness is None:
        parser.error("For --lmm, --kf is required")
    if options.file_covariates is None and \
            options.file_pcs is None and options.regress is True:
        parser.error(("Regress is set to True but neither covariate file",
                      "nor PC file provided"))

    # read data specified in command-line arguments
    dataread = ReadData(verbose=options.verbose)
    dataread.getPhenotypes(file_pheno=options.file_pheno,
                           delim=options.pheno_delim)
    dataread.getGenotypes(
        file_genotypes=options.file_genotypes, delim=options.genotypes_delim)
    if options.file_covariates is not None:
        dataread.getCovariates(file_covariates=options.file_covariates,
                               delim=options.covariate_delim)
    if options.file_relatedness is not None:
        dataread.getRelatedness(file_relatedness=options.file_relatedness,
                                delim=options.relatedness_delim)
    if options.file_cg is not None:
        dataread.getVarianceComponents(file_Cg=options.file_cg,
                                       delim_cg=options.cg_delim,
                                       file_Cn=options.file_cn,
                                       delim_cn=options.cn_delim)
    if options.samplelist is not None or options.file_samplelist is not None:
        samplelist = dataread.getSampleSubset(
            samplelist=options.samplelist,
            file_samplelist=options.file_samplelist)
    else:
        samplelist = None
    if options.traitstring is not None:
        traitlist = dataread.getTraitSubset(traitstring=options.traitstring)
    else:
        traitlist = None

    # combine all input, check for consistency and pre-process data
    datainput = InputData(verbose=options.verbose)
    datainput.addPhenotypes(phenotypes=dataread.phenotypes,
                            phenotype_ID=dataread.phenotype_ID,
                            pheno_samples=dataread.pheno_samples)
    datainput.addGenotypes(genotypes=dataread.genotypes,
                           genotypes_info=dataread.genotypes_info)
    if traitlist is not None:
        datainput.subsetTraits(traitlist=traitlist)
    if dataread.relatedness is not None:
        datainput.addRelatedness(relatedness=dataread.relatedness)
    if dataread.covariates is not None:
        datainput.addCovariates(covariates=dataread.covariates)
    if dataread.Cg is not None:
        datainput.addVarianceComponents(Cg=dataread.Cg, Cn=dataread.Cn)
    datainput.commonSamples(samplelist=samplelist)
    if options.regress is not None:
        datainput.regress()
    if options.transform is not None:
        datainput.transform(transform=options.transform)

    # set up variance decomposition via LiMMBo
    gwas = GWAS(datainput=datainput,
                seed=options.seed,
                verbose=options.verbose)

    if options.singletrait:
        mode = 'singletrait'
    else:
        mode = 'multitrait'

    if options.lmm:
        setup = 'lmm'
    else:
        setup = 'lm'

    resultsAssociation = gwas.runAssociationAnalysis(
        setup=setup,
        mode=mode,
        adjustSingleTrait=options.adjustP)

    if options.nrpermutations is not None:
        pvalues_empirical = gwas.computeEmpiricalP(
            nrpermutations=options.nrpermutations)
    else:
        pvalues_empirical = None

    if options.fdr is not None:
        empirical_fdr, empirical_pvalue_dist = gwas.computeFDR(fdr=options.fdr)

    gwas.saveAssociationResults(resultsAssociation, outdir=options.outdir,
                                name=options.name,
                                pvalues_empirical=pvalues_empirical)

    if options.plot:
        if gwas.fdr_empirical is not None:
            thr = gwas.fdr.empirical
        else:
            thr = options.thr

        gwas.manhattanQQ(results=resultsAssociation,
                         colourS=options.colourS,
                         colourNS=options.colourNS,
                         alphaS=options.alphaS,
                         alphaNS=options.alphaNS,
                         thr_plotting=thr,
                         saveTo=options.outdir)


if __name__ == "__main__":
    entry_point()



from os.path import exists


def file_type(filepath):
    r"""
    Determine filetype based on file ending
    Arguments:
        filepath (string):
            'path/2/file/with/fileending
    Returns:
        (string):
            filetype

    Example:

    .. doctest::

    """
    imexts = ['.png', '.bmp', '.jpg', 'jpeg']
    textexts = ['.csv', '.txt']
    if filepath.endswith('.hdf5') or filepath.endswith('.h5'):
        return 'hdf5'
    if any([filepath.endswith(ext) for ext in textexts]):
        return 'delim'
    if filepath.endswith('.grm.raw'):
        return 'grm.raw'
    if filepath.endswith('.npy'):
        return 'npy'
    if _is_bed(filepath):
        return 'bed'
    if _is_gen(filepath):
        return 'gen'
    if any([filepath.endswith(ext) for ext in imexts]):
        return 'image'
    return 'unknown'


def _is_bed(filepath):
    return all([exists(filepath + ext) for ext in ['.bed', '.bim', '.fam']])


def _is_gen(filepath):
    return all([exists(filepath + ext) for ext in ['.gen', '.sample']])



from limmbo.utils.utils import verboseprint
from limmbo.utils.utils import regularize
from limix.mtset import MTSet as MTST

import time


def vd_reml(datainput, iterations=10, verbose=True):
    r"""
    Compute variance decomposition of phenotypes into genetic and noise
    covariance via standard REML: approach implemented in LIMIX

    Arguments:
        datainput (:class:`InputData`):
            object with ID-matched [`N` x `P`] phenotypes with [`N`]
            individuals and [`P`] phenotypes and [`N` x `N`] relatedness
            estimates
        output (string, optional):
            output directory with user-writing permissions; needed if
            caching is True
        cache (bool, optional):
            should results be cached
        verbose (bool, optional):
            should messages be printed to stdout

    Returns:
        (tuple):
            tuple containing:

            - **Cg** (numpy.array):
              [`P` x `P`] genetic variance component
            - **Cn** (numpy.array):
              [`P` x `P`] noise variance component
            - **process_time** (double):
              cpu time of variance decomposition

    Examples:

    .. doctest::

        >>> import numpy
        >>> from numpy.random import RandomState
        >>> from numpy.linalg import cholesky as chol
        >>> from limmbo.core.vdsimple import vd_reml
        >>> from limmbo.io.input import InputData
        >>> random = RandomState(15)
        >>> N = 100
        >>> S = 1000
        >>> P = 3
        >>> snps = (random.rand(N, S) < 0.2).astype(float)
        >>> kinship = numpy.dot(snps, snps.T) / float(10)
        >>> y  = random.randn(N, P)
        >>> pheno = numpy.dot(chol(kinship), y)
        >>> pheno_ID = [ 'PID{}'.format(x+1) for x in range(P)]
        >>> samples = [ 'SID{}'.format(x+1) for x in range(N)]
            >>> datainput = InputData()
        >>> datainput.addPhenotypes(phenotypes = pheno,
        ...                         phenotype_ID = pheno_ID,
        ...                         pheno_samples = samples)
        >>> datainput.addRelatedness(relatedness = kinship,
        ...                          relatedness_samples = samples)
        >>> Cg, Cn, ptime = vd_reml(datainput, verbose=False)
        >>> Cg.shape
        (3, 3)
    """

    verboseprint(
        "Estimate covariance matrices based on standard REML", verbose=verbose)

    # time variance decomposition
    t0 = time.clock()
    vd = MTST(Y=datainput.phenotypes, R=datainput.relatedness)
    vd_result = vd.fitNull(n_times=iterations, rewrite=True)
    t1 = time.clock()

    if vd_result['conv']:
        verboseprint(
            "Variance decomposition via REML converged", verbose=verbose)
        Cg = vd_result['Cg']
        Cn = vd_result['Cn']
    else:
        verboseprint(
            "Variance decomposition via REML did not converge",
            verbose=verbose)
    processtime = t1 - t0

    # ensure that matrices are true positive semi-definite matrices
    Cg, Cg_ev_min = regularize(Cg, verbose=verbose)
    Cn, Cn_ev_min = regularize(Cn, verbose=verbose)

    return Cg, Cn, processtime



r"""
Data input
**********

Parse command-line arguments
^^^^^^^^^^^^^^^^^^^^^^^^^^^^
.. autofunction:: limmbo.io.parser.getGWASargs

A detailed list of the command-line options for `getGWASargs` can be found in
:ref:`runGWAS`.

.. autofunction:: limmbo.io.parser.getVarianceEstimationArgs

A detailed list of the command-line options for `getVarianceEstimationArgs` can
be found in :ref:`runVarianceEstimation`.

Read data
^^^^^^^^^
.. autoclass:: limmbo.io.reader.ReadData
  :members:

Check data
^^^^^^^^^^
.. autoclass:: limmbo.io.input.InputData
  :members:
"""


from __future__ import absolute_import
from . import parser
from . import reader
from . import input

__all__ = ['parser', 'reader', 'input']



import scipy as sp
import pandas as pd
import numpy as np
import re

from limix.io import read_plink

from limmbo.utils.utils import verboseprint
from limmbo.io.utils import file_type


class MissingInput(Exception):
    """Raised when no appropriate input is given"""
    pass


class FormatError(Exception):
    """Raised when no appropriate input is given"""
    pass


class DataMismatch(Exception):
    """Raised when dimensions of sample/ID names do not match dimension of
     corresponding data"""
    pass


class ReadData(object):
    r"""
    Generate object containing all datasets relevant for the analysis.
    For variance decomposition, at least phenotypes and relatedness estimates
    need to be specified.
    For association testing with LMM, at least phenotype, relatedness estimates
    and genotypes need to be read.
    """

    def __init__(self, verbose=True):
        self.verbose = verbose
        self.samples = None
        self.phenotypes = None
        self.pheno_samples = None
        self.phenotype_ID = None
        self.covariates = None
        self.covs_samples = None
        self.relatedness = None
        self.relatedness_samples = None
        self.genotypes = None
        self.geno_samples = None
        self.genotypes_info = None
        self.pcs = None
        self.pc_samples = None
        self.snps = None
        self.geno_samples = None
        self.position = None
        self.Cg = None
        self.Cn = None

    def getPhenotypes(self, file_pheno=None, delim=","):
        r"""
        Reads [`N` x `P`] phenotype file; file ending must be either .txt or
        .csv

        Arguments:
            file_pheno (string):
                path to [(`N`+1) x (`P`+1)] phenotype file with: [`N`] sample
                IDs in the first column and [`P`] phenotype IDs in the first
                row
            delim (string):
                delimiter of phenotype file, one of " ", ",", "\t"

        Returns:
            None:
                updated the following attributes of the ReadData instance:

                - **self.phenotypes** (np.array):
                  [`N` x `P`] phenotype matrix

        Examples:

            .. doctest::

                >>> from pkg_resources import resource_filename
                >>> from limmbo.io.reader import ReadData
                >>> from limmbo.io.utils import file_type
                >>> data = ReadData(verbose=False)
                >>> file_pheno = resource_filename('limmbo',
                ...                                'io/test/data/pheno.csv')
                >>> data.getPhenotypes(file_pheno=file_pheno)
                >>> data.phenotypes.index[:3]
                Index([u'ID_1', u'ID_2', u'ID_3'], dtype='object')
                >>> data.phenotypes.columns[:3]
                Index([u'trait_1', u'trait_2', u'trait_3'], dtype='object')
        """

        if file_pheno is None:
            raise MissingInput('No phenotype file specified')
        else:
            if file_type(file_pheno) is not "delim":
                raise FormatError('Supplied phenotype file is neither .csv '
                                  'nor .txt')
            verboseprint(
                "Reading phenotypes from %s" % file_pheno,
                verbose=self.verbose)
            try:
                self.phenotypes = pd.io.parsers.read_csv(
                    file_pheno, index_col=0, sep=delim)
            except Exception:
                raise IOError('{} could not be opened'.format(file_pheno))

    def getCovariates(self, file_covariates=None, delim=','):
        r"""
        Reads [`N` x `K`] covariate matrix with [`N`] samples and [`K`]
        covariates.

        Arguments:
            file_covariates (string):
                [`N` x (`K` +1)] covariates file with [`N`] sample IDs in the
                first column
            delim (string):
                delimiter of covariates file, one of " ", ",", "\t"

        Returns:
            None:
                updated the following attributes of the ReadData instance:

                - **self.covariates** (np.array):
                  [`N` x `K`] covariates matrix

        Examples:

            .. doctest::

                >>> from pkg_resources import resource_filename
                >>> from limmbo.io.reader import ReadData
                >>> from limmbo.io.utils import file_type
                >>> data = ReadData(verbose=False)
                >>> file_covs = resource_filename('limmbo',
                ...                               'io/test/data/covs.csv')
                >>> data.getCovariates(file_covariates=file_covs)
                >>> data.covariates.index[:3]
                Index([u'ID_1', u'ID_2', u'ID_3'], dtype='object')
        """

        if file_covariates is not None:
            if file_type(file_covariates) is not 'delim':
                raise FormatError(
                    'Supplied covariate file is not .csv or .txt')
            try:
                self.covariates = pd.io.parsers.read_csv(file_covariates,
                                                         sep=delim,
                                                         index_col=0)
                verboseprint("Reading covariates file", verbose=self.verbose)
            except Exception:
                raise IOError('{} could not be opened'.format(file_covariates))
            # append column of 1's to adjust for mean of covariates
            self.covariates = pd.concat([self.covariates, pd.DataFrame(sp.ones(
                self.covariates.shape[0]), index=self.covariates.index)],
                axis=1)
        else:
            verboseprint("No covariates set", verbose=self.verbose)
            self.covariates = None

    def getRelatedness(self, file_relatedness, delim=","):
        """
        Read file of [`N` x `N`] pairwise relatedness estimates of [`N`]
        samples.

        Arguments:
            file_relatedness (string):
                [(`N` + `1`) x N] .csv file with: [`N`] sample IDs in the first
                row
            delim (string):
                delimiter of covariate file, one of " ", ",", "\t"

        Returns:
            None:
                updated the following attributes of the ReadData instance:

                - **self.relatedness** (np.array):
                  [`N` x `N`] relatedness matrix
        Examples:

            .. doctest::

                >>> from pkg_resources import resource_filename
                >>> from limmbo.io.reader import ReadData
                >>> from limmbo.io.utils import file_type
                >>> data = ReadData(verbose=False)
                >>> file_relatedness = resource_filename('limmbo',
                ...                     'io/test/data/relatedness.csv')
                >>> data.getRelatedness(file_relatedness=file_relatedness)
                >>> data.relatedness.index[:3]
                Index([u'ID_1', u'ID_2', u'ID_3'], dtype='object')
                >>> data.relatedness.columns[:3]
                Index([u'ID_1', u'ID_2', u'ID_3'], dtype='object')
        """

        if file_relatedness is None:
            raise MissingInput('No relatedness data specified')
        if file_type(file_relatedness) is not 'delim':
            raise FormatError('Supplied relatedness file is not .csv or .txt')
        try:
            self.relatedness = pd.io.parsers.read_csv(file_relatedness,
                                                      sep=delim)
            self.relatedness.index = self.relatedness.columns
        except Exception:
            raise IOError('{} could not be opened'.format(file_relatedness))
        verboseprint("Reading relationship matrix", verbose=self.verbose)

    def getPCs(self, file_pcs=None, nrpcs=None, delim=","):
        r"""
        Reads file with [`N` x `PC`] matrix of [`PC`] principal components from
        the genotypes of [`N`] samples.

        Arguments:
            file_pcs (string):
                [`N` x (`PC` +1)] PCA file with [`N`] sample IDs in the first
                column
            delim (string):
                delimiter of PCA file, one of " ", ",", "\t"
            nrpcs (integer):
                Number of PCs to use (uses first nrpcs  principal components)

        Returns:
            None:
                updated the following attributes of the ReadData instance:

                - **self.pcs** (np.array):
                  [`N` x `PC`] principal component matrix

        Examples:

            .. doctest::

                >>> from pkg_resources import resource_filename
                >>> from limmbo.io.reader import ReadData
                >>> from limmbo.io.utils import file_type
                >>> data = ReadData(verbose=False)
                >>> file_pcs = resource_filename('limmbo',
                ...                     'io/test/data/pcs.csv')
                >>> data.getPCs(file_pcs=file_pcs, nrpcs=10, delim=" ")
                >>> data.pcs.index[:3]
                Index([u'ID_1', u'ID_2', u'ID_3'], dtype='object', name=0)
                >>> data.pcs.values[:3,:3]
                array([[-0.02632738, -0.05791269, -0.03619099],
                       [ 0.00761785,  0.00538956,  0.00624196],
                       [ 0.01069307, -0.0205066 ,  0.02299996]])
        """

        if file_type(file_pcs) is not 'delim':
            raise FormatError('Supplied PCA file is not .csv or .txt')

        if file_pcs is not None:
            verboseprint("Reading PCs", verbose=self.verbose)
            self.pcs = pd.io.parsers.read_csv(file_pcs, index_col=0,
                                              sep=delim, header=None)
            if nrpcs is not None:
                verboseprint(
                    "Extracting first %s pcs" % nrpcs, verbose=self.verbose)
                self.pcs = self.pcs.iloc[:, :nrpcs]
        else:
            verboseprint("No pcs set", verbose=self.verbose)
            self.pcs = None

    def getGenotypes(self, file_genotypes=None, delim=','):
        r"""
        Reads genotype file in the following formats: plink (.bed, .bim, .fam),
        gen (.gen, .sample) or comma-separated values (.csv) file.

        Arguments:
            file_geno (string):
                path to phenotype file in .plink or .csv format
                - **plink format**:

                  as specified in the plink `user manual <https://www.cog-
                  genomics.org/plink/1.9/input>`_, binary plink format with
                  .bed, .fam and .bim file.

                - **.csv format**:

                  - [(`NrSNP` + 1) x (`N`+1)] .csv file with: [`N`] sample IDs
                    in the first row and [`NrSNP`] genotype IDs in the first
                    column
                  - sample IDs should be of type: chrom-pos-rsID for instance
                    22-50714616-rs6010226
            delim (string):
                delimiter of genotype file (when text format), one of " ", ",",
                "\t"

        Returns:
            None:
                updated the following attributes of the ReadData instance:

                - **self.genotypes** (np.array):
                  [`N` x `NrSNPs`] genotype matrix
                - **self.genotypes_info** (pd.dataframe):
                  [`NrSNPs` x 2] dataframe with columns 'chrom' and 'pos', and
                  rsIDs as index

        Examples:

            .. doctest::

                >>> from pkg_resources import resource_filename
                >>> from limmbo.io import reader
                >>> from limmbo.io.utils import file_type
                >>> data = reader.ReadData(verbose=False)
                >>> # Read genotypes in delim-format
                >>> file_geno = resource_filename('limmbo',
                ...     'io/test/data/genotypes.csv')
                >>> data.getGenotypes(file_genotypes=file_geno)
                >>> data.genotypes.index[:4]
                Index([u'ID_1', u'ID_2', u'ID_3', u'ID_4'], dtype='object')
                >>> data.genotypes.shape
                (1000, 20)
                >>> data.genotypes_info[:5]
                           chrom       pos
                rs1601111      3  88905003
                rs13270638     8  20286021
                rs75132935     8  76564608
                rs72668606     8  79733124
                rs55770986     7   2087823
                >>> ### read genotypes in plink format
                >>> file_geno = resource_filename('limmbo',
                ...     'io/test/data/genotypes')
                >>> data.getGenotypes(file_genotypes=file_geno)
                >>> data.genotypes_info[:5]
                           chrom       pos
                rs1601111      3  88905003
                rs13270638     8  20286021
                rs75132935     8  76564608
                rs72668606     8  79733124
                rs55770986     7   2087823
        """

        if file_genotypes is None:
            raise MissingInput('No genotypes data specified')
        if file_type(file_genotypes) not in ['delim', 'bed']:
            raise FormatError(('Supplied genotype file is neither in plink '
                               'nor .csv/.txt format'))
        verboseprint("Reading genotypes from %s" % file_genotypes,
                     verbose=self.verbose)

        if file_type(file_genotypes) is 'delim':
            try:
                genotypes = pd.io.parsers.read_csv(
                    file_genotypes, sep=delim, index_col=0, header=0)
            except Exception:
                raise IOError('{} could not be opened'.format(file_genotypes))

            info = np.array(genotypes.index)

            genotypes_info = []
            snp_ID = []
            for id in range(info.shape[0]):
                split = np.array(info[id].split('-'))
                snp_ID.append(split[2])
                genotypes_info.append(split[[0, 1]])

            self.genotypes = genotypes.astype(float).T
            self.genotypes_info = pd.DataFrame(
                np.array(genotypes_info),
                columns=['chrom', 'pos'],
                index=snp_ID)

        if file_type(file_genotypes) is 'bed':
            try:
                (bim, fam, bed) = read_plink(file_genotypes,
                                             verbose=self.verbose)
            except Exception:
                raise IOError('{} could not be opened'.format(file_genotypes))

            self.genotypes = pd.DataFrame(bed.compute()).astype(float).T
            self.genotypes.index = fam.iid
            self.genotypes_info = pd.DataFrame(
                np.array([bim.chrom, bim.pos]).T,
                columns=['chrom', 'pos'],
                index=bim.snp)
            self.genotypes_info.index.name = None

    def getVarianceComponents(self, file_Cg=None, file_Cn=None, delim_cg=",",
                              delim_cn=","):
        r"""
        Reads a comma-separated files with [`P` x `P`] matrices of [`P`] trait
        covariance estimates.

        Arguments:
            file_Cg (string):
                [`P` x `P`] .csv file with [`P`] trait covariance estimates of
                the genetic component
            file_Cn (string):
                [`P` x `P`] .csv file with [`P`] trait covariance estimates of
                the non-genetic (noise) component

        Returns:
            None:
                updated the following attributes of the ReadData instance:

                - **self.Cg** (np.array):
                  [`P` x `P`] matrix with trait covariance of the genetic
                  component
                - **self.Cn** (np.array):
                  [`P` x `P`] matrix with trait covariance of the non-genetic
                  (noise) component

        Examples:

            .. doctest::

                >>> from pkg_resources import resource_filename
                >>> from limmbo.io.reader import ReadData
                >>> data = ReadData()
                >>> file_Cg = resource_filename(
                ...     'limmbo', 'io/test/data/Cg.csv')
                >>> file_Cn = resource_filename(
                ...     'limmbo', 'io/test/data/Cn.csv')
                >>> data.getVarianceComponents(file_Cg=file_Cg,
                ...                            file_Cn=file_Cn)
                >>> data.Cg.shape
                (10, 10)
                >>> data.Cn.shape
                (10, 10)
                >>> data.Cg[:3,:3]
                array([[ 0.45446454, -0.21084613,  0.01440468],
                       [-0.21084613,  0.11443656,  0.01250233],
                       [ 0.01440468,  0.01250233,  0.02347906]])
                >>> data.Cn[:3,:3]
                array([[ 0.53654803, -0.14392748, -0.45483001],
                       [-0.14392748,  0.88793093,  0.30539822],
                       [-0.45483001,  0.30539822,  0.97785614]])
        """

        if file_Cg is None and file_Cn is None:
            verboseprint(
                ("No variance components supplied, estimate variance "
                 "components before lmm test"), verbose=self.verbose)
        elif file_Cg is None or file_Cn is None:
            raise IOError('Both variant components need to be supplied:',
                          'Cg is %s and Cn is %s') % (file_Cg, file_Cn)
        else:
            try:
                self.Cg = np.array(
                    pd.io.parsers.read_csv(file_Cg, header=None, sep=delim_cg))
            except Exception:
                raise IOError('{} could not be opened'.format(file_Cg))
            try:
                self.Cn = np.array(
                    pd.io.parsers.read_csv(file_Cn, header=None, sep=delim_cn))
            except Exception:
                raise IOError('{} could not be opened'.format(file_Cn))

    def getTraitSubset(self, traitstring=None):
        """
        Limit analysis to specific subset of traits

        Arguments:
            traitstring (string):
                comma-separated trait numbers (for single traits) or hyphen-
                separated trait numbers (for trait ranges) or combination of
                both for trait selection (1-based)

        Returns:
            (numpy array)
                array containing list of trait IDs

        Examples:

            .. doctest::

                >>> from limmbo.io import reader
                >>> data = reader.ReadData(verbose=False)
                >>> traitlist = data.getTraitSubset("1,3,5,7-10")
                >>> print traitlist
                [0 2 4 6 7 8 9]
        """
        if traitstring is None:
            verboseprint('No trait subset chosen', verbose=self.verbose)
        else:
            verboseprint('Chose subset of {} traits'.format(traitstring),
                         verbose=self.verbose)
            search = re.compile('[^0-9,-]').search
            if bool(search(traitstring)):
                raise FormatError(('Traitstring can only contain integers '
                                   '(0-9), comma (,) and hyphen (-), but {}'
                                   'provided').format(traitstring))
            traitslist = [x.split('-') for x in traitstring.split(',')]
            traitsarray = []
            for t in traitslist:
                if len(t) == 1:
                    traitsarray.append(int(t[0]) - 1)
                else:
                    [traitsarray.append(x)
                     for x in range(int(t[0]) - 1, int(t[1]))]
            return np.array(traitsarray)

    def getSampleSubset(self, file_samplelist=None, samplelist=None):
        r"""
        Read file or string with subset of sample IDs to analyse.

        Arguments:
            file_samplelist (string):
                "path/to/file_samplelist": file contains subset sample IDs with
                one ID per line, no header.
            samplestring (string):
                comma-separated list of sample IDs e.g. "ID1,ID2,ID5,ID10".

        Returns:
            (numpy array)
                array containing list of sample IDs
        """
        if file_samplelist is not None and samplelist is not None:
            raise IOError("Only one of file_samplelist or samplelist can "
                          "be specified")
        if file_samplelist is not None or samplelist is not None:
            if file_samplelist is not None:
                verboseprint("Read sample list from file",
                             verbose=self.verbose)
                try:
                    samplelist = pd.io.parsers.read_csv(
                        file_samplelist, sep=" ", header=None, index_col=0)
                except Exception:
                    raise IOError('{} could not be opened'.format(
                        file_samplelist))
                samplelist = samplelist.index
            else:
                verboseprint("Split sample string", verbose=self.verbose)
                samplelist = samplelist.split(",")

            verboseprint("Number of samples in sample list: %s" %
                         len(samplelist), verbose=self.verbose)
            return samplelist



import argparse


def getGWASargs():
    parser = argparse.ArgumentParser(
        prog="runAssociation",
        description=('Models the association between phenotypes and '
                     'genotypes, accepting additional covariates and '
                     'parameters to account for population structure and '
                     'relatedness between samples. Users can choose between '
                     'single-trait and multi-trait models, simple linear or '
                     'linear mixed model set-ups.'))
    required = parser.add_argument_group('Basic required arguments')
    required.add_argument(
        '-p',
        '--file_pheno',
        action="store",
        dest="file_pheno",
        required=False,
        help=('Path [string] to [(N+1) x (P+1)] .csv file of [P] '
              'phenotypes with [N] samples (first column: sample IDs, first '
              'row: phenotype IDs). Default: %(default)s'))
    required.add_argument(
        '--pheno_delim',
        action="store",
        dest="pheno_delim",
        required=False,
        default=",",
        help=('Delimiter of phenotype file.  g '
              'Default: %(default)s'))
    required.add_argument(
        '-g',
        '--file_geno',
        action="store",
        dest="file_genotypes",
        required=False,
        default=None,
        help=('Genotype file: either [S x N].csv file (first column: SNP id, '
              'first row: sample IDs) or  `plink formated <https://www.cog-'
              'genomics.org/plink/1.9/input>`_ genotypes '
              '(.bim/.fam/.bed). Default: %(default)s'))
    required.add_argument(
        '--geno_delim',
        action="store",
        dest="genotypes_delim",
        required=False,
        default=",",
        help=('Delimiter of genotype file (if not in `plink format <https://'
              'www.cog-genomics.org/plink/1.9/input>`_). '
              'Default: %(default)s'))
    required.add_argument(
        '-o',
        '--outdir',
        action="store",
        dest="outdir",
        required=True,
        help=('Path [string] of output directory; user needs writing '
              'permission. Default: %(default)s'))
    mode = required.add_mutually_exclusive_group(required=True)
    mode.add_argument(
        '-st',
        '--singletrait',
        action="store_true",
        dest="singletrait",
        default=False,
        help=('Set flag to conduct a single-trait association analyses'
              'Default: %(default)s'))

    mode.add_argument(
        '-mt',
        '--multitrait',
        action="store_true",
        dest="multitrait",
        default=False,
        help=('Set flag to conduct a multi-trait association analyses'
              'Default: %(default)s'))

    setup = required.add_mutually_exclusive_group(required=True)
    setup.add_argument(
        '-lm',
        '--lm',
        action="store_true",
        dest="lm",
        default=False,
        help=('Set flag to use a simple linear model for the association'
              'analysis'))
    setup.add_argument(
        '-lmm',
        '--lmm',
        action="store_true",
        dest="lmm",
        default=False,
        help=('Set flag to use a linear mixed model for the association'
              'analysis'))

    output = parser.add_argument_group('Output arguments')
    output.add_argument(
        '-n',
        '--name',
        action="store",
        dest="name",
        default="",
        help=('Name (used for output file naming). Default: '
              '%(default)s'))

    output.add_argument(
        '-v',
        '--verbose',
        action="store_true",
        dest="verbose",
        required=False,
        default=False,
        help=('[bool]: should analysis progress be displayed. '
              'Default: %(default)s'))

    optionalfiles = parser.add_argument_group('Optional files')
    optionalfiles.add_argument(
        '-k',
        '--file_kinship',
        action="store",
        dest="file_relatedness",
        required=False,
        default=None,
        help=('Path [string] to [N x (N+1)] file of kinship/relatedness '
              'matrix with [N] samples (first row: sample IDs). Required when '
              '--lmm/-lm. Default: '
              '%(default)s'))
    optionalfiles.add_argument(
        '--kinship_delim',
        action="store",
        dest="relatedness_delim",
        required=False,
        default=",",
        help=('Delimiter of kinship file. '
              'Default: %(default)s'))
    optionalfiles.add_argument(
        '-cg',
        '--file_cg',
        action="store",
        dest="file_cg",
        required=False,
        default=None,
        help=('Required for large phenotype sizes when --lmm/-lm; computed '
              'via runLiMMBo; specifies file name for genetic trait covariance'
              ' matrix (rows: traits, columns: traits). Default: %(default)s'))
    optionalfiles.add_argument(
        '--cg_delim',
        action="store",
        dest="cg_delim",
        required=False,
        default=",",
        help=('Delimiter of Cg file. Default: %(default)s'))
    optionalfiles.add_argument(
        '-cn',
        '--file_cn',
        action="store",
        dest="file_cn",
        required=False,
        default=None,
        help=('Required for large phenotype sizeswhen --lmm/-lm; computed via '
              'runLiMMBo; specifies file name for non-genetic trait '
              'covariance matrix (rows: traits, columns: traits). '
              'Default: %(default)s'))
    optionalfiles.add_argument(
        '--cn_delim',
        action="store",
        dest="cn_delim",
        required=False,
        default=",",
        help=('Delimiter of Cn file. g Default: %(default)s'))
    optionalfiles.add_argument(
        '-pcs',
        '--file_pcs',
        action="store",
        dest="file_pcs",
        required=False,
        default=None,
        help=('Path to [N x PCs] file of principal components from '
              'genotypes to be included as covariates (first column: '
              'sample IDs, first row: PC IDs); Default: %(default)s'))
    optionalfiles.add_argument(
        '--pcs_delim',
        action="store",
        dest="pcs_delim",
        required=False,
        default=",",
        help=('Delimiter of PCs file. g Default: %(default)s'))
    optionalfiles.add_argument(
        '-c',
        '--file_cov',
        action="store",
        dest="file_covariates",
        required=False,
        default=None,
        help=('Path [string] to [(N+1) x C] file of covariates matrix with '
              '[N] samples and [K] covariates (first column: sample IDs, '
              'first row: phenotype IDs). Default: %(default)s'))
    optionalfiles.add_argument(
        '--covariate_delim',
        action="store",
        dest="covariate_delim",
        required=False,
        default=",",
        help=('Delimiter of covariates file. g '
              'Default: %(default)s'))

    optional = parser.add_argument_group('Optional association parameters')
    optional.add_argument(
        '-adjustP',
        '--adjustP',
        action="store",
        dest="adjustP",
        choices=['bonferroni', 'effective', 'None'],
        default=None,
        required=False,
        type=str,
        help=('Method to adjust single-trait p-values for'
              'multiple hypothesis testing when running'
              'multiple single-trait GWAS: bonferroni/effective number of '
              'tests `(Galwey,2009) <http://onlinelibrary.wiley.com/doi/10.'
              '1002/gepi.20408/abstract>`_ . Default: %(default)s'))

    optional.add_argument(
        '-nrpermutations',
        '--nrpermutations',
        action="store",
        dest="nrpermutations",
        default=None,
        required=False,
        type=int,
        help=('Number of permutations for computing empirical p-values; '
              '1/nrpermutations is maximum level of testing for '
              'significance. Default: %(default)s')
    )
    optional.add_argument(
        '-fdr',
        '--fdr',
        action="store",
        dest="fdr",
        required=False,
        default=None,
        type=float,
        help=('FDR threshold for computing empirical FDR. Default: '
              '%(default)s'))
    optional.add_argument(
        '-seed',
        '--seed',
        action="store",
        dest="seed",
        default=256,
        required=False,
        help=('Seed [int] to inittiate random number generation for '
              'permutations. Default: %(default)s'),
        type=int)

    data = parser.add_argument_group('Optional data processing parameters')
    data.add_argument(
        '-tr',
        '--transform_method',
        action="store",
        dest="transform",
        default='scale',
        choices=['scale', 'gaussian'],
        required=False,
        help=('Choose type [string] of data preprocessing: scale (mean '
              'center, divide by sd) or gaussian (inverse normalise). '
              'Default: %(default)s'))
    data.add_argument(
        '-reg',
        '--reg_covariates',
        action="store_true",
        dest="regress",
        required=False,
        help=('[bool]: should covariates be regressed out? Default: '
              '%(default)s'))

    subset = parser.add_argument_group('Optional subsetting options')
    subset.add_argument(
        '-traitset',
        '--traitset',
        action="store",
        dest="traitstring",
        required=False,
        default=None,
        help=('Comma- (for list of traits) or hyphen- (for trait range) or '
              'comma and hyphen-separated list [string] of traits (trait '
              'columns) to choose; default: None (=all traits). Default: '
              '%(default)s'))

    subset.add_argument(
        '-nrpcs',
        '--nrpcs',
        action="store",
        dest="nrpcs",
        required=False,
        default=10,
        help='First PCs to chose. Default: %(default)s',
        type=int)

    samples = subset.add_mutually_exclusive_group(required=False)
    samples.add_argument(
        '--file_samplelist',
        action="store",
        dest="file_samplelist",
        required=False,
        default=None,
        help=('Path [string] to file with samplelist for sample '
              'selection, with one sample ID per line. Default: %(default)s'))

    samples.add_argument(
        '--samplelist',
        action="store",
        dest="samplelist",
        required=False,
        default=None,
        help=('Comma-separated list [string] of samples IDs to restrict '
              'analysis to, e.g. ID1,ID2,ID5,ID9,ID10. Default: %(default)s'))

    plotting = parser.add_argument_group('Plot arguments',
                                         'Arguments for depicting GWAS '
                                         'results as manhattan plot')
    plotting.add_argument(
        '--plot',
        action="store_true",
        dest="plot",
        required=False,
        default=False,
        help=('Set flag if results of association analysis should be depicted '
              'as manhattan and quantile-quantile plot'
              'Default: %(default)s'))
    plotting.add_argument(
        '-colourS',
        '--colourS',
        action="store",
        dest="colourS",
        required=False,
        default='DarkBlue',
        help=('Colour of significant points in manhattan plot'
              'Default: %(default)s'))
    plotting.add_argument(
        '-colourNS',
        '--colourNS',
        action="store",
        dest="colourNS",
        required=False,
        default='Orange',
        help=('Colour of non-significant points in manhattan plot'
              'Default: %(default)s'))
    plotting.add_argument(
        '-alphaS',
        '--alphaS',
        action="store",
        dest="alphaS",
        required=False,
        default=1.0,
        type=float,
        help=('Transparency of significant points in manhattan plot'
              'Default: %(default)s'))
    plotting.add_argument(
        '-alphaNS',
        '--alphaNS',
        action="store",
        dest="alphaNS",
        required=False,
        default=0.1,
        type=float,
        help=('Transparency of non-significant points in manhattan plot'
              'Default: %(default)s'))
    plotting.add_argument(
        '-thr',
        '--threshold',
        action="store",
        dest="thr",
        required=False,
        default=None,
        help=('Significance threshold; when --fdr specified, empirical fdr '
              'used as threshold. Default: %(default)s'))

    version = parser.add_argument_group('Version')
    version.add_argument('--version', action='version',
                         version='%(prog)s 0.1.3')

    return parser


def getVarianceEstimationArgs():
    parser = argparse.ArgumentParser(
        prog="runVarianceEstimation",
        description=('Estimates the genetic and non-genetic trait'
                     'covariance matrix parameters of a linear mixed model '
                     'with random genetic and non-genetic effect via a '
                     'bootstrapping-based approach.'))
    required = parser.add_argument_group('Basic required arguments')
    vd = required.add_mutually_exclusive_group(required=True)
    vd.add_argument(
        '--limmbo',
        action="store_true",
        dest="limmbo",
        required=False,
        default=False,
        help=('Set to run covariance estimation with LiMMBo, recommended for '
              'trait sizes larger than 20; Default: %(default)s'))
    vd.add_argument(
        '--reml',
        action="store_true",
        dest="reml",
        required=False,
        default=False,
        help=('Set to run covariance estimation with simple reml, recommended '
              'for trait sizes less than 20. Default: %(default)s'))
    required.add_argument(
        '-p',
        '--file_pheno',
        action="store",
        dest="file_pheno",
        required=False,
        help=('Path [string] to [(N+1) x (P+1)] .csv file of [P] '
              'phenotypes with [N] samples (first column: sample IDs, first '
              'row: phenotype IDs). Default: %(default)s'))
    required.add_argument(
        '--pheno_delim',
        action="store",
        dest="pheno_delim",
        required=False,
        default=",",
        help=('Delimiter of phenotype file. '
              'Default: %(default)s'))
    required.add_argument(
        '-k',
        '--file_kinship',
        action="store",
        dest="file_relatedness",
        required=False,
        default=None,
        help=('Path [string] to [N x (N+1)] file of kinship/relatedness '
              'matrix with [N] samples (first row: sample IDs). Required when '
              '--lmm/-lm. Default: '
              '%(default)s'))
    required.add_argument(
        '--kinship_delim',
        action="store",
        dest="relatedness_delim",
        required=False,
        default=",",
        help=('Delimiter of kinship file. '
              'Default: %(default)s'))
    required.add_argument(
        '-o',
        '--outdir',
        action="store",
        dest="outdir",
        required=True,
        help=('Path [string] of output directory; user needs writing '
              'permission. Default: %(default)s'))

    optionalfiles = parser.add_argument_group('Optional files')
    optionalfiles.add_argument(
        '-c',
        '--file_cov',
        action="store",
        dest="file_covariates",
        required=False,
        default=None,
        help=('Path [string] to [(N+1) x C] file of covariates matrix with '
              '[N] samples and [K] covariates (first column: sample IDs, '
              'first row: phenotype IDs). Default: %(default)s'))
    optionalfiles.add_argument(
        '--covariate_delim',
        action="store",
        dest="covariate_delim",
        required=False,
        default=",",
        help=('Delimiter of covariates file. '
              'Default: %(default)s'))

    limmbo = parser.add_argument_group('Bootstrapping parameters')
    limmbo.add_argument(
        '-seed',
        '--seed',
        action="store",
        dest="seed",
        default=234,
        required=False,
        help=('seed [int] used to generate bootstrap matrix. Default: '
              '%(default)s'),
        type=int)

    limmbo.add_argument(
        '-sp',
        '--smallp',
        action="store",
        dest="S",
        default=None,
        required=True,
        help=('Size [int] of phenotype subsamples used for variance '
              'decomposition. Default: %(default)s'),
        type=int)
    limmbo.add_argument(
        '-r',
        '--runs',
        action="store",
        dest="runs",
        default=None,
        required=False,
        help='Total number [int] of bootstrap runs. Default: %(default)s',
        type=int)
    limmbo.add_argument(
        '-t',
        '--timing',
        action="store_true",
        dest="timing",
        default=False,
        required=False,
        help=('[bool]: should variance decomposition be timed. Default: '
              '%(default)s'))
    limmbo.add_argument(
        '--minCooccurrence',
        action="store",
        dest="minCooccurrence",
        default=3,
        type=int,
        required=False,
        help=('Minimum count [int] of the pairwise sampling of any given '
              'trait pair. Default: %(default)s'))
    limmbo.add_argument(
        '-i',
        '--iterations',
        action="store",
        dest="iterations",
        default=10,
        required=False,
        type=int,
        help=('Number [int] of iterations for variance decomposition '
              'attempts. Default: %(default)s'))
    limmbo.add_argument(
        '-cpus',
        '--cpus',
        action="store",
        dest="cpus",
        default=None,
        required=False,
        type=int,
        help=('Number [int] of available CPUs for parallelisation of '
              'variance decomposition steps. Default: %(default)s'))

    data = parser.add_argument_group('Optional data processing parameters')
    data.add_argument(
        '-tr',
        '--transform_method',
        action="store",
        dest="transform",
        default='scale',
        choices=['scale', 'gaussian'],
        required=False,
        help=('Choose type [string] of data preprocessing: scale (mean '
              'center, divide by sd) or gaussian (inverse normalise). '
              'Default: %(default)s'))

    subset = parser.add_argument_group('Optional subsetting options')
    subset.add_argument(
        '-traitset',
        '--traitset',
        action="store",
        dest="traitstring",
        required=False,
        default=None,
        help=('Comma- (for list of traits) or hyphen- (for trait range) or '
              'comma and hyphen-separated list [string] of traits (trait '
              'columns) to choose; default: None (=all traits). Default: '
              '%(default)s'))

    samples = subset.add_mutually_exclusive_group(required=False)
    samples.add_argument(
        '--file_samplelist',
        action="store",
        dest="file_samplelist",
        required=False,
        default=None,
        help=('Path [string] to file with samplelist for sample '
              'selection, with one sample ID per line.Default: %(default)s'))

    samples.add_argument(
        '--samplelist',
        action="store",
        dest="samplelist",
        required=False,
        default=None,
        help=('Comma-separated list [string] of samples IDs to restrict '
              'analysis to, e.g. ID1,ID2,ID5,ID9,ID10. Default: %(default)s'))

    output = parser.add_argument_group('Output arguments')
    output.add_argument(
        '-dontSaveIntermediate',
        '--dontSaveIntermediate',
        action="store_false",
        dest="intermediate",
        default=True,
        required=False,
        help=('Set to suppress saving intermediate variance components. '
              'Default: %(default)s'))

    output.add_argument(
        '-v',
        '--verbose',
        action="store_true",
        dest="verbose",
        required=False,
        default=False,
        help=('[bool]: should analysis step description be printed. '
              'Default: %(default)s'))

    version = parser.add_argument_group('Version')
    version.add_argument('--version', action='version',
                         version='%(prog)s 0.1.3')

    return parser



from limix.util.preprocess import regressOut

from limmbo.utils.utils import verboseprint
from limmbo.utils.utils import scale
from limmbo.utils.utils import AlleleFrequencies

import pandas as pd
import numpy as np

from scipy_sugar.stats import quantile_gaussianize
from math import sqrt


class MissingInput(Exception):
    """Raised when no appropriate input is given"""
    pass


class FormatError(Exception):
    """Raised when no appropriate input is given"""
    pass


class DataMismatch(Exception):
    """Raised when dimensions of sample/ID names do not match dimension of
    corresponding data"""
    pass


class InputData(object):
    """
    Generate object containing all datasets relevant for variance decomposition
    (phenotypes, relatedness estimates) and pre-processing steps (check for
    common samples and sample order, covariates regression and phenotype
    transformation)

    Arguments:
        verbose (bool):
            initialise verbose: should progress messages be printed to stdout
    """

    def __init__(self, verbose=True):
        self.verbose = verbose
        self.samples = None
        self.phenotypes = None
        self.pheno_samples = None
        self.phenotype_ID = None
        self.covariates = None
        self.covariate_samples = None
        self.relatedness = None
        self.relatedness_samples = None
        self.pcs = None
        self.pc_samples = None
        self.snps = None
        self.genotypes = None
        self.geno_samples = None
        self.genotypes_info = None
        self.Cg = None
        self.Cn = None

    def addPhenotypes(self, phenotypes, pheno_samples=None, phenotype_ID=None):
        """
        Add phenotypes, their phenotype ID and their sample IDs to
        InputData instance

        Arguments:
            phenotypes (array-like):
                [`N x `P`] phenotype matrix of `N` individuals and `P`
                phenotypes; if pandas.DataFrame with pheno_samples as index and
                phenotypes_ID as columns, pheno_samples and phenotype_ID do not
                have to specified separately.
            pheno_samples (array-like):
                [`N`] sample ID
            phenotype_ID (array-like):
                [`P`] phenotype IDs

        Returns:
            None:
                updated the following attributes of the InputData instance:

                - **self.phenotypes** (pd.DataFrame):
                  [`N` x `P`] phenotype array
                - **self.pheno_samples** (np.array):
                  [`N`] sample IDs
                - **self.phenotype_ID** (np.array):
                  [`P`] phenotype IDs


        Examples:

            .. doctest::

                >>> from limmbo.io import input
                >>> import numpy as np
                >>> import pandas as pd
                >>> pheno = np.array(((1,2),(7,1),(3,4)))
                >>> pheno_samples = ['S1','S2', 'S3']
                >>> phenotype_ID = ['ID1','ID2']
                >>> phenotypes = pd.DataFrame(pheno, index=pheno_samples,
                ...     columns = phenotype_ID)
                >>> indata = input.InputData(verbose=False)
                >>> indata.addPhenotypes(phenotypes = phenotypes)
                >>> print indata.phenotypes.shape
                (3, 2)
                >>> print indata.pheno_samples.shape
                (3,)
                >>> print indata.phenotype_ID.shape
                (2,)
        """
        if pheno_samples is None:
            try:
                self.pheno_samples = np.array(phenotypes.index)
            except Exception:
                raise TypeError(("pheno_samples are not provided and "
                                 "phenotypes has no index to retrieve "
                                 "pheno_samples from."))
        else:
            self.pheno_samples = np.array(pheno_samples)

        if phenotype_ID is None:
            try:
                self.phenotype_ID = np.array(phenotypes.columns)
            except Exception:
                raise TypeError(("phenotype_ID are not provided and "
                                 "phenotypes has no column names to retrieve "
                                 "phenotype_ID from."))
        else:
            self.phenotype_ID = np.array(phenotype_ID)

        if phenotypes.shape[0] != self.pheno_samples.shape[0]:
            raise DataMismatch(('Number of samples in phenotypes ({}) does '
                                'not match number of sample IDs ({}) provided'
                                ).format(
                phenotypes.shape[0], self.pheno_samples.shape[0]))
        if phenotypes.shape[1] != self.phenotype_ID.shape[0]:
            raise DataMismatch(('Number phenotypes ({}) does not match '
                                'number of phenotype IDs ({}) provided'
                                ).format(phenotypes.shape[1],
                                         self.phenotype_ID.shape[0]))
        if len(self.pheno_samples) != len(set(self.pheno_samples)):
            raise IOError("Duplicate sample names in phenotypes")
        if len(self.phenotype_ID) != len(set(self.phenotype_ID)):
            raise IOError("Duplicate trait names in phenotypes")

        self.phenotypes = pd.DataFrame(phenotypes, index=self.pheno_samples,
                                       columns=self.phenotype_ID)

    def addCovariates(self, covariates, covs_samples=None):
        """
        Add [`N` x `K`] covariate data with [`N`] samples and [`K`] covariates
        to InputData instance.

        Arguments:
            covariates (array-like):
                [`N x `K`] covariate matrix of `N` individuals and `K`
                covariates; if pandas.DataFrame with covs_samples as index,
                covs_samples do not have to specified separately.
            covs_samples (array-like):
                [`N`] sample ID

        Returns:
            None:
                updated the following attributes of the InputData instance:

                - **self.covariates** (pd.DataFrame):
                  [`N` x `K`] covariates matrix
                - **self.covs_samples** (np.array):
                  [`N`] sample IDs

        Examples:

            .. doctest::

                >>> from limmbo.io import input
                >>> import numpy as np
                >>> import pandas as pd
                >>> covariates = [(1,2,4),(1,1,6),(0,4,8)]
                >>> covs_samples = ['S1','S2', 'S3']
                >>> covariates = pd.DataFrame(covariates, index=covs_samples)
                >>> indata = input.InputData(verbose=False)
                >>> indata.addCovariates(covariates = covariates,
                ...     covs_samples = covs_samples)
                >>> print indata.covariates.shape
                (3, 3)
                >>> print indata.covs_samples.shape
                (3,)

        """
        if covs_samples is None:
            try:
                self.covs_samples = np.array(covariates.index)
            except Exception:
                raise TypeError(("covs_samples are not provided and "
                                 "covariates has no index to retrieve "
                                 "covs_samples from."))
        else:
            self.covs_samples = np.array(covs_samples)
        if np.array(covariates).shape[0] != np.array(self.covs_samples
                                                     ).shape[0]:
            raise DataMismatch(('Number of samples in covariates ({}) does '
                                'not match number of sample IDs ({}) provided'
                                ).format(
                np.array(covariates).shape[0],
                np.array(self.covs_samples).shape[0]))
        if len(self.covs_samples) != len(set(self.covs_samples)):
            raise IOError("Duplicate sample names in covariates")
        self.covariates = pd.DataFrame(covariates, index=self.covs_samples)

    def addRelatedness(self, relatedness, relatedness_samples=None):
        """
        Add [`N` x `N`] pairwise relatedness estimates of [`N`] samples to the
        InputData instance

        Arguments:
            relatedness (array-like):
                [`N x `N`] relatedness matrix of `N` individuals;
                if pandas.DataFrame with relatedness_samples as index,
                relatedness_samples do not have to specified separately.
            relatedness_samples (array-like):
                [`N`] sample IDs

        Returns:
            None:
                updated the following attributes of the InputData instance:

                - **self.relatedness** (pd.DataFrame):
                  [`N` x `N`] relatedness matrix
                - **self.relatedness_samples** (np.array):
                  [`N`] sample IDs

        Examples:

            .. doctest::

                >>> from limmbo.io import input
                >>> import numpy
                >>> import pandas as pd
                >>> from numpy.random import RandomState
                >>> from numpy.linalg import cholesky as chol
                >>> random = RandomState(5)
                >>> N = 100
                >>> SNP = 1000
                >>> X = (random.rand(N, SNP) < 0.3).astype(float)
                >>> relatedness = numpy.dot(X, X.T)/float(SNP)
                >>> relatedness_samples = numpy.array(
                ...     ['S{}'.format(x+1) for x in range(N)])
                >>> relatedness = pd.DataFrame(relatedness,
                ...     index=relatedness_samples)
                >>> indata = input.InputData(verbose=False)
                >>> indata.addRelatedness(relatedness = relatedness)
                >>> print indata.relatedness.shape
                (100, 100)
                >>> print indata.relatedness_samples.shape
                (100,)

        """
        if relatedness_samples is None:
            try:
                self.relatedness_samples = np.array(relatedness.index)
            except Exception:
                raise TypeError(("relatedness_samples are not provided and "
                                 "relatedness has no index to retrieve "
                                 "relatedness_samples from"))
        else:
            self.relatedness_samples = np.array(relatedness_samples)
        rel = np.array(relatedness)
        if rel.shape[0] != rel.shape[1]:
            raise FormatError(('Relatedness has to be a square matrix, but '
                               'number of rows {} is not equal to number of '
                               'columns {}').format(
                                   rel.shape[0],
                                   rel.shape[1]))

        if not np.all(np.array(rel) - np.array(rel).T == 0):
            raise FormatError('Relatedness matrix is not symmetric')
        if not self._is_positive_definite(rel):
            raise FormatError(
                'Relatedness matrix is not positive-semi definite')
        if rel.shape[0] != self.relatedness_samples.shape[0]:
            raise DataMismatch(('Number of samples in relatedness ({}) does '
                                'not match number of sample IDs ({}) provided'
                                ).format(
                rel.shape[0], self.relatedness_samples.shape[0]))
        if len(self.relatedness_samples) != len(set(self.relatedness_samples)):
            raise IOError("Duplicate sample names in relatedness")
        self.relatedness = pd.DataFrame(relatedness,
                                        index=self.relatedness_samples,
                                        columns=self.relatedness_samples)

    def addGenotypes(self, genotypes, geno_samples=None,
                     genotypes_info=None):
        """
        Add [`N` x `NrSNP`] genotype array of [`N`] samples and [`NrSNP`]
        genotypes, [`N`] array of sample IDs and [`NrSNP` x 2] dataframe of
        genotype description to InputData instance.

        Arguments:
            genotypes (array-like):
                [`N` x `NrSNP`] genotype array of [`N`] samples and [`NrSNP`]
                genotypes; if pandas.DataFrame with geno_samples as index,
                geno_samples do not have to specified separately.
            geno_samples (array-like):
                [`N`] vector of `N` sample IDs
            genotypes_info (dataframe):
                  [`NrSNPs` x 2] dataframe with columns 'chrom' and 'pos', and
                  rsIDs as index

        Returns:
            None:
                updated the following attributes of the InputData instance:

                - **self.genotypes** (pd.DataFrame):
                  [`N` x `NrSNPs`] genotype matrix
                - **self.geno_samples** (np.array):
                  [`N`] sample IDs
                - **self.genotypes_info** (pd.DataFrame):
                  [`NrSNPs` x 2] dataframe with columns 'chrom' and 'pos', and
                  rsIDs as index

        Examples:

            .. doctest::

                >>> from pkg_resources import resource_filename
                >>> from limmbo.io import reader
                >>> from limmbo.io import input
                >>> data = reader.ReadData(verbose=False)
                >>> file_geno = resource_filename(
                ...     'limmbo', 'io/test/data/genotypes.csv')
                >>> data.getGenotypes(file_genotypes=file_geno)
                >>> indata = input.InputData(verbose=False)
                >>> indata.addGenotypes(genotypes=data.genotypes,
                ...                     genotypes_info=data.genotypes_info)
                >>> indata.geno_samples[:5]
                array(['ID_1', 'ID_2', 'ID_3', 'ID_4', 'ID_5'], dtype=object)
                >>> indata.genotypes.shape
                (1000, 20)
                >>> indata.genotypes_info[:5]
                           chrom       pos
                rs1601111      3  88905003
                rs13270638     8  20286021
                rs75132935     8  76564608
                rs72668606     8  79733124
                rs55770986     7   2087823
        """
        if geno_samples is None:
            try:
                self.geno_samples = np.array(genotypes.index)
            except Exception:
                raise TypeError(("geno_samples are not provided and genotypes "
                                 "has no index to retrieve geno_samples from"))
        else:
            self.geno_samples = np.array(geno_samples)
        if genotypes_info is None:
            raise MissingInput(('Genotype info has to be specified via '
                                'genotypes_info'))
        self.genotypes = pd.DataFrame(genotypes, index=self.geno_samples)
        self.genotypes_info = genotypes_info
        if self.genotypes.shape[0] != self.geno_samples.shape[0]:
            raise DataMismatch(('Number of samples in genotypes ({}) does '
                                'not match number of sample IDs ({}) provided'
                                ).format(
                self.genotypes.shape[0], self.geno_samples.shape[0]))
        if self.genotypes.shape[1] != self.genotypes_info.shape[0]:
            raise DataMismatch(('Number of genotypes in genotypes ({}) does '
                                'not match number of genotypes in '
                                'genotypes_info ({})').format(
                self.genotypes.shape[1], self.genotypes_info.shape[0]))
        if len(self.geno_samples) != len(set(self.geno_samples)):
            raise IOError("Duplicate sample names in genotypes")

    def addVarianceComponents(self, Cg, Cn,):
        """
        Add [`P` x `P`] matrices of [`P`] trait covariance estimates
        of the genetic trait variance component (Cg) and the non-genetic
        (noise) variance component (Cn) to InputData instance.

        Arguments:
            Cg (array-like):
                [`P x `P`] matrix of `P` trait covariance estimates of the
                genetic trait covaraince component
            Cn (array-like):
                [`P x `P`] matrix of `P` trait covariance estimates of the
                non-genetic (noise) trait covaraince component

        Returns:
            None:
                updated the following attributes of the InputData instance:

                - **self.Cg** (np.array):
                  [`P x `P`] matrix of `P` trait covariance estimates of the
                  genetic trait covariance component
                - **self.Cn** (np.array):
                  [`P x `P`] matrix of `P` trait covariance estimates of the
                  non-genetic trait covaraince component

        Examples:

            .. doctest::

                >>> from pkg_resources import resource_filename
                >>> from limmbo.io import reader
                >>> from limmbo.io import input
                >>> import numpy as np
                >>> from numpy.random import RandomState
                >>> from numpy.linalg import cholesky as chol
                >>> data = reader.ReadData(verbose=False)
                >>> file_pheno = resource_filename('limmbo',
                ...                     'io/test/data/pheno.csv')
                >>> data.getPhenotypes(file_pheno=file_pheno)
                >>> file_Cg = resource_filename('limmbo',
                ...                     'io/test/data/Cg.csv')
                >>> file_Cn = resource_filename('limmbo',
                ...                     'io/test/data/Cn.csv')
                >>> data.getVarianceComponents(file_Cg=file_Cg,
                ...                            file_Cn=file_Cn)
                >>> indata = input.InputData(verbose=False)
                >>> indata.addPhenotypes(phenotypes = data.phenotypes)
                >>> indata.addVarianceComponents(Cg = data.Cg, Cn=data.Cn)
                >>> print indata.Cg.shape
                (10, 10)
                >>> print indata.Cg.shape
                (10, 10)
        """
        if Cg is not None and Cn is not None:
            if self.phenotypes is None:
                raise FormatError(('Phenotypes have to be added before Cg/Cn '
                                   'can be added'))
            self.Cg = np.array(Cg)
            self.Cn = np.array(Cn)
            if self.Cg.shape[0] != self.Cg.shape[1]:
                raise FormatError(('Cg has to be a square matrix, but '
                                   'number of rows {} is not equal to number '
                                   'of columns {}').format(self.Cg.shape[0],
                                                           self.Cg.shape[1]))
            if not np.all(self.Cg - self.Cg.T == 0):
                raise FormatError('Cg is not symmetric')
            if not self._is_positive_definite(self.Cg):
                raise FormatError('Cg is not positive-semi definite')
            if self.Cg.shape[0] != self.phenotypes.shape[1]:
                raise DataMismatch(('Number of traits in Cg ({}) does '
                                    'not match number of traits ({}) in '
                                    'phenotypes'
                                    ).format(self.Cg.shape[0],
                                             self.phenotypes.shape[1]))
            if self.Cn.shape[0] != self.Cn.shape[1]:
                raise FormatError(('Cn has to be a square matrix, but '
                                   'number of rows {} is not equal to number '
                                   'of columns {}').format(self.Cn.shape[0],
                                                           self.Cn.shape[1]))

            if not np.all(self.Cn - self.Cn.T == 0):
                raise FormatError('Cn is not symmetric')
            if not self._is_positive_definite(self.Cn):
                raise FormatError('Cn is not positive-semi definite')
            if self.Cn.shape[0] != self.phenotypes.shape[1]:
                raise DataMismatch(('Number of traits in Cn ({}) does '
                                    'not match number of traits ({}) in '
                                    'phenotypes').format(
                                        self.Cn.shape[0],
                                        self.phenotypes.shape[1])
                                   )

    def addPCs(self, pcs, pc_samples=None):
        """
        Add [`N` x `PC`] matrix of [`PC`] principal components from the
        genotypes of [`N`] samples to InputData instance.

        Arguments:
            pcs (array-like):
                [`N x `PCs`] principal component matrix of `N` individuals and
                `PCs` principal components; if pandas.DataFrame with pc_samples
                as index, covs_samples do not have to specified separately.
            pc_samples (array-like):
                [`N`] sample IDs

        Returns:
            None:
                updated the following attributes of the InputData instance:

                - **self.pcs** (pd.DataFrame):
                  [`N` x `PCs`] principal component matrix
                - **self.pc_samples** (np.array):
                  [`N`] sample IDs

        Examples:

            .. doctest::

                >>> from pkg_resources import resource_filename
                >>> from limmbo.io import reader
                >>> from limmbo.io import input
                >>> data = reader.ReadData(verbose=False)
                >>> file_pcs = resource_filename('limmbo',
                ...                     'io/test/data/pcs.csv')
                >>> data.getPCs(file_pcs=file_pcs, nrpcs=10, delim=" ")
                >>> indata = input.InputData(verbose=False)
                >>> indata.addPCs(pcs = data.pcs)
                >>> print indata.pcs.shape
                (1000, 10)
                >>> print indata.pc_samples.shape
                (1000,)

        """
        if pc_samples is None:
            try:
                self.pc_samples = np.array(pcs.index)
            except Exception:
                raise TypeError(("pc_samples are not provided and pcs has "
                                 "no index to retrieve pc_samples from"))
        else:
            self.pc_samples = np.array(pc_samples)
        if np.array(pcs).shape[0] != np.array(self.pc_samples).shape[0]:
            raise DataMismatch(('Number of samples in pcs ({}) does'
                                'not match number of sample IDs ({}) provided'
                                ).format(
                np.array(pcs).shape[0],
                np.array(pc_samples).shape[0]))
        if len(self.pc_samples) != len(set(self.pc_samples)):
            raise IOError("Duplicate sample names for principle components")
        self.pcs = pd.DataFrame(pcs, index=self.pc_samples)

    def subsetTraits(self, traitlist=None):
        """
        Limit analysis to specific subset of traits

        Arguments:
            traitlist (array-like):
                array of trait numbers to select from phenotypes

        Returns:
            None:
                updated the following attributes of the InputData instance:

                - **self.traitlist** (list):
                  of [`t`] trait numbers (int) to choose for analysis
                - **self.phenotypes** (pd.DataFrame):
                  reduced set of [`N` x `t`] phenotypes
                - **self.phenotype.ID** (np.array):
                  reduced set of [`t`] phenotype IDs

        Examples:

            .. doctest::

                >>> from pkg_resources import resource_filename
                >>> from limmbo.io.reader import ReadData
                >>> from limmbo.io.input import InputData
                >>> from limmbo.io.utils import file_type
                >>> data = ReadData(verbose=False)
                >>> file_pheno = resource_filename('limmbo',
                ...                                'io/test/data/pheno.csv')
                >>> data.getPhenotypes(file_pheno=file_pheno)
                >>> traitlist = data.getTraitSubset(traitstring="1-3,5")
                >>> indata = InputData(verbose=False)
                >>> indata.addPhenotypes(phenotypes = data.phenotypes)
                >>> print indata.phenotypes.shape
                (1000, 10)
                >>> print indata.phenotype_ID.shape
                (10,)
                >>> indata.subsetTraits(traitlist=traitlist)
                >>> print indata.phenotypes.shape
                (1000, 4)
                >>> print indata.phenotype_ID.shape
                (4,)
        """
        self.traitlist = np.array(traitlist)
        if len(self.traitlist) != len(set(self.traitlist)):
            raise IOError("Duplicate trait names in traitlist")
        try:
            self.phenotypes = self.phenotypes.iloc[:, self.traitlist]
            self.phenotype_ID = self.phenotype_ID[self.traitlist]
        except:
            raise DataMismatch(('Selected trait number {} is greater '
                                'than number of phenotypes provided {}'
                                ).format(max(self.traitlist) + 1,
                                         self.phenotypes.shape[1]))

    def commonSamples(self, samplelist=None):
        """
        Get [`M]` common samples out of phenotype, relatedness and optional
        covariates with [`N`] samples (if all samples present in all datasets
        [`M`] = [`N`]) and ensure that samples are in same order.

        Arguments:
            samplelist (array-like):
                array of sample IDs to select from data

        Returns:
            None:
                updated the following attributes of the InputData instance:

                - **self.phenotypes** (pd.DataFrame):
                  [`M` x `P`] phenotype matrix
                - **self.pheno_samples** (np.array):
                  [`M`] sample IDs
                - **self.relatedness** (pd.DataFrame):
                  [`M x M`] relatedness matrix
                - **self.relatedness_samples** (np.array):
                  [`M`] sample IDs of relatedness matrix
                - **self.covariates** (pd.DataFrame):
                  [`M` x `K`] covariates matrix
                - **self.covs_samples** (np.array):
                  [`M`] sample IDs
                - **self.genotypes** (pd.DataFrame):
                  [`M` x `NrSNPs`] genotypes matrix
                - **self.geno_samples** (np.array):
                  [`M`] sample IDs
                - **self.pcs** (pd.DataFrame):
                  [`M` x `PCs`] principal component matrix
                - **self.pc_samples** (np.array):
                  [`M`] sample IDs

        Examples:

            .. doctest::

                >>> from limmbo.io import input
                >>> import numpy as np
                >>> import pandas as pd
                >>> from numpy.random import RandomState
                >>> from numpy.linalg import cholesky as chol
                >>> random = RandomState(5)
                >>> P = 2
                >>> K = 4
                >>> N = 10
                >>> SNP = 1000
                >>> pheno = random.normal(0,1, (N, P))
                >>> pheno_samples = np.array(['S{}'.format(x+4)
                ...     for x in range(N)])
                >>> phenotype_ID = np.array(['ID{}'.format(x+1)
                ...     for x in range(P)])
                >>> phenotypes = pd.DataFrame(pheno, index=pheno_samples,
                ...     columns=phenotype_ID)
                >>> X = (random.rand(N, SNP) < 0.3).astype(float)
                >>> relatedness = np.dot(X, X.T)/float(SNP)
                >>> relatedness_samples = np.array(['S{}'.format(x+1)
                ...     for x in range(N)])
                >>> covariates = random.normal(0,1, (N-2, K))
                >>> covs_samples = np.array(['S{}'.format(x+1)
                ...     for x in range(N-2)])
                >>> indata = input.InputData(verbose=False)
                >>> indata.addPhenotypes(phenotypes = pheno,
                ...                      pheno_samples = pheno_samples,
                ...                      phenotype_ID = phenotype_ID)
                >>> indata.addRelatedness(relatedness = relatedness,
                ...                  relatedness_samples = relatedness_samples)
                >>> indata.addCovariates(covariates = covariates,
                ...                      covs_samples = covs_samples)
                >>> indata.covariates.shape
                (8, 4)
                >>> indata.phenotypes.shape
                (10, 2)
                >>> indata.relatedness.shape
                (10, 10)
                >>> indata.commonSamples(samplelist=["S4", "S6", "S5"])
                >>> indata.covariates.shape
                (3, 4)
                >>> indata.phenotypes.shape
                (3, 2)
                >>> indata.relatedness.shape
                (3, 3)
        """
        self.samples = self.pheno_samples

        if self.relatedness is not None:
            test_pheno_relatedness = np.intersect1d(self.pheno_samples,
                                                    self.relatedness_samples)
            if len(test_pheno_relatedness) == 0:
                raise DataMismatch(('No common samples between phenotypes and '
                                    'relatedness estimates'))
            self.samples = test_pheno_relatedness
        if self.genotypes is not None:
            test_pheno_geno = np.intersect1d(self.pheno_samples,
                                             self.geno_samples)
            if len(test_pheno_geno) == 0:
                raise DataMismatch(('No common samples between phenotypes,'
                                    'and genotypes'))
            self.samples = np.intersect1d(self.samples, test_pheno_geno)
        if self.covariates is not None:
            test_pheno_covs = np.intersect1d(self.pheno_samples,
                                             self.covs_samples)
            if len(test_pheno_covs) == 0:
                raise DataMismatch(('No common samples between phenotypes, '
                                    'and covariates'))
            self.samples = np.intersect1d(self.samples, test_pheno_covs)
        if self.pcs is not None:
            test_pheno_pcs = np.intersect1d(self.pheno_samples,
                                            self.pcs_samples)
            if len(test_pheno_pcs) == 0:
                raise DataMismatch(('No common samples between phenotypes,'
                                    'and pcs'))
            self.samples = np.intersect1d(self.samples, test_pheno_pcs)

        if samplelist is not None:
            if len(samplelist) != len(set(samplelist)):
                raise IOError("Duplicate sample names in samplelist")
            test_samples_samplelist = np.intersect1d(self.samples, samplelist)
            if len(test_samples_samplelist) == 0:
                raise DataMismatch(('No samples between common samples in, '
                                    'datasets and samplelist'))
            if len(test_samples_samplelist) < len(samplelist):
                raise DataMismatch(('Not all Ids in samplelist are contained '
                                    'in common samples from provided '
                                    'datasets'))
            self.samples = samplelist

        self.phenotypes = self.phenotypes.loc[self.samples, :]
        self.pheno_samples = np.array(self.phenotypes.index)

        if self.genotypes is not None:
            self.genotypes = self.genotypes.loc[self.samples, :]
            self.geno_samples = np.array(self.genotypes.index)
        if self.relatedness is not None:
            self.relatedness = self.relatedness.loc[self.samples, :]
            self.relatedness = self.relatedness[self.samples]
            self.relatedness_samples = np.array(self.relatedness.index)
        if self.covariates is not None:
            self.covariates = self.covariates.loc[self.samples, :]
            self.covs_samples = np.array(self.covariates.index)
        if self.pcs is not None:
            self.pcs = self.pcs.loc[self.samples, :]
            self.pc_samples = np.array(self.pcs.index)

    def regress(self):
        """
        Regress out covariates (optional).

        Returns:
            None:
                updated the following attributes of the InputData instance:

                - **self.phenotypes** (np.array):
                  [`M` x `P`] phenotype matrix of residuals of linear model
                - **self.covariates**:
                  None

        Examples:

            .. doctest::

                >>> from limmbo.io import input
                >>> import numpy as np
                >>> from numpy.random import RandomState
                >>> from numpy.linalg import cholesky as chol
                >>> random = RandomState(5)
                >>> P = 5
                >>> K = 4
                >>> N = 100
                >>> pheno = random.normal(0,1, (N, P))
                >>> pheno_samples = np.array(['S{}'.format(x+1)
                ...     for x in range(N)])
                >>> phenotype_ID = np.array(['ID{}'.format(x+1)
                ...     for x in range(P)])
                >>> covariates = random.normal(0,1, (N, K))
                >>> covs_samples = np.array(['S{}'.format(x+1)
                ...     for x in range(N)])
                >>> indata = input.InputData(verbose=False)
                >>> indata.addPhenotypes(phenotypes = pheno,
                ...                      pheno_samples = pheno_samples,
                ...                      phenotype_ID = phenotype_ID)
                >>> indata.addCovariates(covariates = covariates,
                ...                      covs_samples = covs_samples)
                >>> indata.phenotypes.values[:3, :3]
                array([[ 0.44122749, -0.33087015,  2.43077119],
                       [ 1.58248112, -0.9092324 , -0.59163666],
                       [-1.19276461, -0.20487651, -0.35882895]])
                >>> indata.regress()
                >>> indata.phenotypes.values[:3, :3]
                array([[ 0.34421705, -0.01470998,  2.25710966],
                       [ 1.69886647, -1.41756814, -0.55614649],
                       [-1.10700674, -0.66017713, -0.22201814]])
        """
        if np.array_equal(self.phenotypes, self.covariates):
            raise DataMismatch(('Phenotype and covariate arrays are '
                                'identical'))
        verboseprint('Regress covariates', verbose=self.verbose)
        phenotypes = regressOut(np.array(self.phenotypes),
                                np.array(self.covariates))
        self.phenotypes = pd.DataFrame(phenotypes,
                                       index=self.phenotypes.index,
                                       columns=self.phenotypes.columns)
        self.covariates = None

    def transform(self, transform):
        """
        Transform phenotypes

        Arguments:
            transform (string):
                transformation method for phenotype data:

                    - scale:
                      mean center, divide by sd
                    - gaussian:
                      inverse normalisation

        Returns:
            None:
                updated the following attributes of the InputData instance:

                - **self.phenotypes** (np.array):
                  [`N` x `P`] (transformed) phenotype matrix

        Examples:

            .. doctest::

                >>> from limmbo.io import input
                >>> import numpy as np
                >>> from numpy.random import RandomState
                >>> from numpy.linalg import cholesky as chol
                >>> random = RandomState(5)
                >>> P = 5
                >>> K = 4
                >>> N = 100
                >>> pheno = random.normal(0,1, (N, P))
                >>> pheno_samples = np.array(['S{}'.format(x+1)
                ...     for x in range(N)])
                >>> phenotype_ID = np.array(['ID{}'.format(x+1)
                ...     for x in range(P)])
                >>> SNP = 1000
                >>> X = (random.rand(N, SNP) < 0.3).astype(float)
                >>> relatedness = np.dot(X, X.T)/float(SNP)
                >>> relatedness_samples = np.array(['S{}'.format(x+1)
                ...     for x in range(N)])
                >>> indata = input.InputData(verbose=False)
                >>> indata.addPhenotypes(phenotypes = pheno,
                ...                      pheno_samples = pheno_samples,
                ...                      phenotype_ID = phenotype_ID)
                >>> indata.addRelatedness(relatedness = relatedness,
                ...                  relatedness_samples = relatedness_samples)
                >>> indata.phenotypes.values[:3, :3]
                array([[ 0.44122749, -0.33087015,  2.43077119],
                       [ 1.58248112, -0.9092324 , -0.59163666],
                       [-1.19276461, -0.20487651, -0.35882895]])
                >>> indata.transform(transform='gaussian')
                >>> indata.phenotypes.values[:3, :3]
                array([[ 0.23799988, -0.11191464,  2.05785598],
                       [ 1.41041953, -0.81365681, -0.92217818],
                       [-1.55977999,  0.01240937, -0.62091817]])
        """
        if transform == 'scale':
            verboseprint('Use %s as transformation' % transform,
                         verbose=self.verbose)
            phenotypes = scale(self.phenotypes)
            self.phenotypes = pd.DataFrame(phenotypes,
                                           index=self.phenotypes.index,
                                           columns=self.phenotypes.columns)
        elif transform == 'gaussian':
            verboseprint('Use %s as transformation' % transform,
                         verbose=self.verbose)
            phenotypes = np.apply_along_axis(quantile_gaussianize, 0,
                                             self.phenotypes)
            self.phenotypes = pd.DataFrame(phenotypes,
                                           index=self.phenotypes.index,
                                           columns=self.phenotypes.columns)
        else:
            raise TypeError(('Possible transformation methods are: scale, '
                             'and gaussian but {} provided').format(transform))

    def standardiseGenotypes(self):
        r"""
        Standardise genotypes:

        .. math::
           w_{ij} = \frac{x_{ij} -2p_i}{\sqrt{2p_i (1-p_i)}}

        where :math:`x_{ij}` is the number of copies of the reference allele
        for the :math:`i` th SNP of the :math:`j` th individual and :math:`p_i`
        is the frequency of the reference allele (as described in `(Yang et al
        2011) <http://www.pubmedcentral.nih.gov/articlerender.fcgi?artid=
        3014363&tool=pmcentrez&rendertype=abstract>`_).

        Returns:
            None:
                updated the following attributes of the InputData instance:

                - **self.genotypes_sd** (numpy array):
                  [`N` x `NrSNP`] matrix of `NrSNP` standardised genotypes for
                  `N` samples.

        Examples:

            .. doctest::

                >>> from pkg_resources import resource_filename
                >>> from limmbo.io import reader
                >>> from limmbo.io import input
                >>> from limmbo.utils.utils import makeHardCalledGenotypes
                >>> from limmbo.utils.utils import AlleleFrequencies
                >>> data = reader.ReadData(verbose=False)
                >>> file_geno = resource_filename(
                ...     'limmbo', 'io/test/data/genotypes.csv')
                >>> data.getGenotypes(file_genotypes=file_geno)
                >>> indata = input.InputData(verbose=False)
                >>> indata.addGenotypes(genotypes=data.genotypes,
                ...                     genotypes_info=data.genotypes_info)
                >>> geno_sd = indata.standardiseGenotypes()
                >>> geno_sd.iloc[:5,:3]
                             0         1       2
                ID_1 -2.201123 -2.141970 -8.9622
                ID_2 -2.201123 -2.141970 -8.9622
                ID_3 -2.201123 -2.141970 -8.9622
                ID_4  0.908627 -0.604125 -8.9622
                ID_5 -0.646248 -2.141970 -8.9622
        """
        self.genotypes_sd = np.zeros(self.genotypes.shape)
        for snp in range(self.genotypes.shape[1]):
            p, q = AlleleFrequencies(self.genotypes.iloc[:, snp])
            var_snp = sqrt(2 * p * q)
            for n in range(self.genotypes.iloc[:, snp].shape[0]):
                self.genotypes_sd[n, snp] = (np.array(self.genotypes)[n, snp] -
                                             2 * q) / var_snp

        self.genotypes_sd = pd.DataFrame(self.genotypes_sd,
                                         index=self.genotypes.index)

        return self.genotypes_sd

    def getAlleleFrequencies(self):
        """
        Compute allele frequencies of genotypes.

        Returns:
            None:
                updated the following attributes of the InputData instance:

                - **self.freqs** (pandas DataFrame):
                  [`NrSNP` x `2`] matrix of alt and ref allele frequencies;
                  index: snp IDs.

        Examples:

            .. doctest::

                >>> from pkg_resources import resource_filename
                >>> from limmbo.io import reader
                >>> from limmbo.io import input
                >>> from limmbo.utils.utils import makeHardCalledGenotypes
                >>> from limmbo.utils.utils import AlleleFrequencies
                >>> data = reader.ReadData(verbose=False)
                >>> file_geno = resource_filename(
                ...     'limmbo', 'io/test/data/genotypes.csv')
                >>> data.getGenotypes(file_genotypes=file_geno)
                >>> indata = input.InputData(verbose=False)
                >>> indata.addGenotypes(genotypes=data.genotypes,
                ...                     genotypes_info=data.genotypes_info,
                ...                     geno_samples=data.geno_samples)
                >>> freqs = indata.getAlleleFrequencies()
                >>> freqs.iloc[:5,:]
                                   p         q
                rs1601111   0.292186  0.707814
                rs13270638  0.303581  0.696419
                rs75132935  0.024295  0.975705
                rs72668606  0.119091  0.880909
                rs55770986  0.169338  0.830662
        """
        verboseprint('Get allele frequencies of %s snps'.format(
            self.genotypes.shape[1]), verbose=self.verbose)
        self.freqs = np.zeros((self.genotypes.shape[1], 2))
        for snp in range(self.genotypes.shape[1]):
            self.freqs[snp, 0], self.freqs[snp, 1] = AlleleFrequencies(
                np.array(self.genotypes)[:, snp])

        self.freqs = pd.DataFrame(self.freqs, index=self.genotypes_info.index,
                                  columns=['p', 'q'])
        return self.freqs

    @staticmethod
    def _is_positive_definite(matrix):
        try:
            np.linalg.cholesky(matrix)
            return(True)
        except np.linalg.linalg.LinAlgError:
            return(False)



from __future__ import absolute_import

from .testit import test
from . import core
from . import io
from . import utils
from . import plot
from . import bin

__all__ = ['test', 'core', 'io', 'utils', 'plot', 'bin']
__version__ = "0.1.3"



from limmbo.io.parser import getVarianceEstimationArgs
from limmbo.io.reader import ReadData
from limmbo.io.input import InputData
from limmbo.core.vdbootstrap import LiMMBo
from limmbo.core.vdsimple import vd_reml

import pandas as pd


def entry_point():

    # parse command-line arguments
    parser = getVarianceEstimationArgs()
    options = parser.parse_args()
    if options.file_covariates is None and options.regress is True:
        parser.error(("Regress is set to True but no covariate file",
                      "provided"))
    # read data specified in command-line arguments
    dataread = ReadData(verbose=options.verbose)
    dataread.getPhenotypes(file_pheno=options.file_pheno,
                           delim=options.pheno_delim)
    dataread.getRelatedness(
        file_relatedness=options.file_relatedness,
        delim=options.relatedness_delim)
    if options.file_covariates is not None:
        dataread.getCovariates(file_covariates=options.file_covariates,
                               delim=options.covariate_delim)
    if options.samplelist is not None or options.file_samplelist is not None:
        samplelist = dataread.getSampleSubset(
                samplelist=options.samplelist,
                file_samplelist=options.file_samplelist)
    else:
        samplelist = None
    if options.traitstring is not None:
        traitlist = dataread.getTraitSubset(traitstring=options.traitstring)
    else:
        traitlist = None

    # combine all input, check for consistency and pre-process data
    datainput = InputData(verbose=options.verbose)
    datainput.addPhenotypes(phenotypes=dataread.phenotypes)
    datainput.addRelatedness(relatedness=dataread.relatedness)
    if dataread.covariates is not None:
        datainput.addCovariates(covariates=dataread.covariates)
        datainput.regress()
    datainput.commonSamples(samplelist=samplelist)
    datainput.subsetTraits(traitlist=traitlist)
    if options.transform is not None:
        datainput.transform(transform=options.transform)

    if options.limmbo:
        datalimmbo = LiMMBo(datainput=datainput,
                            S=options.S,
                            timing=options.timing,
                            iterations=options.iterations,
                            verbose=options.verbose)
        resultsBS = datalimmbo.runBootstrapCovarianceEstimation(
            seed=options.seed, cpus=options.cpus,
            minCooccurrence=options.minCooccurrence,
            n=options.runs)

        resultsCovariance = datalimmbo.combineBootstrap(results=resultsBS)
        datalimmbo.saveVarianceComponents(resultsCovariance,
                                          output=options.outdir,
                                          intermediate=options.intermediate)

    if options.reml:
        Cg, Cn, processtime = vd_reml(datainput=datainput,
                                      iterations=options.iterations,
                                      verbose=options.verbose)
        try:
            pd.DataFrame(Cg).to_csv('{}/Cg_REML.csv'.format(
                options.outdir), sep=",", header=False, index=False)
            pd.DataFrame(Cn).to_csv('{}/Cn_REML.csv'.format(
                options.outdir), sep=",", header=False, index=False)
            pd.DataFrame([processtime]).to_csv(
                '{}/process_time_REML.csv'.format(
                    options.outdir), sep=",", header=False, index=False)
        except:
            raise IOError(('Cannot write to {}: check writing '
                           'permissions').format(options.outdir))


if __name__ == "__main__":
    entry_point()



from limmbo.utils.utils import verboseprint
from limmbo.utils.utils import nans
from limmbo.utils.utils import regularize
from limmbo.utils.utils import inflate_matrix

import scipy as sp
from scipy.optimize import fmin_l_bfgs_b as opt
import pandas as pd
import numpy as np
import bottleneck as bn
import time
import cPickle

import limix.mtset

from limix_core.covar import FreeFormCov
import pp


class DataMismatch(Exception):
    """Raised when dimensions of sample/ID names do not match dimension of
    corresponding data"""
    pass


class LiMMBo(object):
    def __init__(self, datainput, S, timing=False, iterations=10,
                 verbose=False):
        r"""
        Arguments:
            datainput ():
                fdas
            timing (bool):
                if set to True, process time will be recorded
            iterations (int):

            S (int):
                subsampling size `S`

        """
        self.phenotypes = datainput.phenotypes
        self.relatedness = datainput.relatedness
        self.S = S
        self.timing = timing
        self.iterations = iterations
        self.verbose = verbose

        try:
            self.phenotypes = np.array(self.phenotypes)
        except:
            raise IOError(
                "datainput.phenotypes cannot be converted to np.array")

        try:
            self.relatedness = np.array(self.relatedness)
        except:
            raise IOError(
                "datainput.relatedness cannot be converted to np.array")

        if self.S > self.phenotypes.shape[1]:
            raise DataMismatch(("Subsampling size S ({}) greater than number "
                                "of phenotypes ({})").format(
                                    self.S,
                                    self.phenotypes.shape[1]))

    def runBootstrapCovarianceEstimation(self, seed, cpus, minCooccurrence=3,
                                         n=None):
        r"""
        Distribute variance decomposition of subset matrices via pp

        Arguments:
            seed (int):
                seed to initialise random number generator for bootstrapping
            minCooccurrence (int):
                minimum number a trait pair should be sampled; once reached
                for all trait pairs, sampling is stopped if n is None;
                default=3
            n (int):
                if not None, sets the total number of permutations,
                otherwise n determined by minCooccurrence;  default: None
            cpus (int):
                number of cpus available for covariance estimation

        Returns:
            (list):
                list containing variance components for all bootstrap runs
        """

        self.P = self.phenotypes.shape[1]
        self.__generateBootstrapMatrix(seed=seed, n=n,
                                       minCooccurrence=minCooccurrence)
        ppservers = ()
        jobs = []
        results = []

        if cpus is not None:
            job_server = pp.Server(cpus, ppservers=ppservers)
        else:
            job_server = pp.Server(ppservers=ppservers)

        verboseprint(
            'Number of CPUs available for parallelising: {}'.format(
                job_server.get_ncpus()),
            verbose=self.verbose)

        for bs in range(self.runs):
            pheno = self.__bootstrapPhenotypes(bs)
            verboseprint('Start vd for bootstrap nr {}'.format(bs + 1))
            jobs.append(
                job_server.submit(self.__VarianceDecomposition, (pheno, bs),
                                  (verboseprint, ), ("limix.mtset", "time")))

        for job in jobs:
            bsresult = job()
            bsresult['bootstrap'] = self.bootstrap_matrix[bsresult[
                'bsindex'], :]
            results.append(bsresult)

        return results

    def combineBootstrap(self, results):
        r"""
        Combine the [`S` x `S`] subset covariance matrices to find the overall
        [`P` x `P`] covariance matrices Cg and Cn.

        Arguments:
            results (list):
                results of runBootstrapVarianceDecomposition()

        Returns:
            (dictionary):
                dictionary containing:

                - **Cg_fit** (numpy.array):
                  [`P` x `P`] genetic covariance matrix via fitting
                - **Cn_fit** (numpy.array):
                  [`P` x `P`] noise covariance matrix via fitting
                - **Cg_average** (numpy.array):
                  [`P` x `P`] genetic covariance matrix via simple average
                - **Cn_average** (numpy.array):
                  [`P` x `P`] noise covariance matrix via simple average
                - **Cg_all_bs** (numpy.array):
                  [`runs` x `S` x `S`] genetic covariance matrices of `runs`
                  phenotype subsets of size `S`
                - **Cn_all_bs** (numpy.array):
                  [`runs` x `S` x `S`] noise covariance matrices of `runs`
                  phenotype subsets of size `S`
                - **proc_time_ind_bs** (list):
                  individual run times for all variance decomposition runs of
                  [`S` x `S`] Cg and Cn
                - **proc_time_sum_ind_bs** (list):
                  sum of individual run times for all variance decomposition
                  runs of [`S` x `S`] Cg and Cn
                - **proc_time_combine_bs** (list):
                  run time for finding [`P` x `P`] trait covariance estimates
                  from fitting [`S` x `S`] bootstrap covariance estimates
                - **nr_of_bs** (int):
                  number of bootstrap runs
                - **nr_of_successful_bs** (int):
                  total number of successful bootstrapping runs i.e. variance
                  decomposition converged
                - **results_fit_Cg** ():
                  results parameters of the bfgs-fit of the genetic covariance
                  matrices (via scipy.optimize.fmin_l_bfgs_g)
                - **results_fit_Cn** ():
                  results parameters of the bfgs-fit of the noise covariance
                  matrices (via scipy.optimize.fmin_l_bfgs_g)
        """

        verboseprint('Combine bootstrapping results...', verbose=self.verbose)
        time0 = time.clock()
        bs_results = self.__getBootstrapResults(results=results)
        time1 = time.clock()

        proc_time_combine_bs = time1 - time0
        proc_time_sum_ind_bs = np.array(bs_results['process_time_bs']).sum()

        verboseprint("Check Cg (average):", verbose=self.verbose)
        Cg_average, Cg_average_ev_min = regularize(bs_results['Cg_average'])
        verboseprint("Check Cn (average):", verbose=self.verbose)
        Cn_average, Cn_average_ev_min = regularize(bs_results['Cn_average'])
        verboseprint("Check Cg (fit):", verbose=self.verbose)
        Cg_fit, Cg_fit_ev_min = regularize(bs_results['Cg_fit'])
        verboseprint("Check Cn (fit):", verbose=self.verbose)
        Cn_fit, Cn_fit_ev_min = regularize(bs_results['Cn_fit'])

        results = {'Cg_fit': Cg_fit, 'Cn_fit': Cn_fit,
                   'Cg_average': Cg_average, 'Cn_average': Cn_average,
                   'Cg_all_bs': bs_results['Cg_bs'],
                   'Cn_all_bs': bs_results['Cn_bs'],
                   'proc_time_ind_bs': bs_results['process_time_bs'],
                   'proc_time_sum_ind_bs': proc_time_sum_ind_bs,
                   'proc_time_combine_bs': proc_time_combine_bs,
                   'nr_bs': self.runs,
                   'nr_successful_bs': bs_results['number_of_successful_bs'],
                   'results_fit_Cg': bs_results['results_fit_Cg'],
                   'results_fit_Cn': bs_results['results_fit_Cn']
                   }

        return results

    def saveVarianceComponents(self, resultsCombineBootstrap, output,
                               intermediate=True):
        r"""
        Save variance components as comma-separated files or python objects
        (via Cpickle).

        Arguments:
            resultsCombineBootstrap (dictionary):

                - **Cg_fit** (numpy.array):
                  [`P` x `P`] genetic covariance matrix via fitting
                - **Cn_fit** (numpy.array):
                  [`P` x `P`] noise covariance matrix via fitting
                - **Cg_average** (numpy.array):
                  [`P` x `P`] genetic covariance matrix via simple average
                - **Cn_average** (numpy.array):
                  [`P` x `P`] noise covariance matrix via simple average
                - **Cg_all_bs** (numpy.array):
                  [`runs` x `S` x `S`] genetic covariance matrices of `runs`
                  phenotype subsets of size `S`
                - **Cn_all_bs** (numpy.array):
                  [`runs` x `S` x `S`] noise covariance matrices of `runs`
                  phenotype subsets of size `S`
                - **proc_time_ind_bs** (list):
                  individual run times for all variance decomposition runs of
                  [`S` x `S`] Cg and Cn
                - **proc_time_sum_ind_bs** (list):
                  sum of individual run times for all variance decomposition
                  runs of [`S` x `S`] Cg and Cn
                - **proc_time_combine_bs** (list):
                  run time for finding [`P` x `P`] trait covariance estimates
                  from fitting [`S` x `S`] bootstrap covariance estimates
                - **nr_of_bs** (int):
                  number of bootstrap runs
                - **nr_of_successful_bs** (int):
                  total number of successful bootstrapping runs i.e. variance
                  decomposition converged
                - **results_fit_Cg** ():
                  results parameters of the bfgs-fit of the genetic covariance
                  matrices (via scipy.optimize.fmin_l_bfgs_g)
                - **results_fit_Cn** ():
                  results parameters of the bfgs-fit of the noise covariance
                  matrices (via scipy.optimize.fmin_l_bfgs_g)

            output (string):
                path/to/directory where variance components will be saved;
                needs writing permission
            intermediate (bool):
                if set to True, intermediate variance components (average
                covariance matrices, bootstrap matrices and results parameter
                of BFGS fit) are saved

        Returns:
            None
        """

        verboseprint("Generate output files", verbose=self.verbose)

        verboseprint("Write [`P` x `P`] covariance matrices",
                     verbose=self.verbose)
        pd.DataFrame(resultsCombineBootstrap['Cg_fit']).to_csv(
            '{}/Cg_fit_seed{}.csv'.format(output, self.seed), sep=",",
            header=False, index=False)
        pd.DataFrame(resultsCombineBootstrap['Cn_fit']).to_csv(
            '{}/Cn_fit_seed{}.csv'.format(output, self.seed), sep=",",
            header=False, index=False)

        if intermediate:
            verboseprint("Write bootstrap matrix", verbose=self.verbose)
            pd.DataFrame(self.bootstrap_matrix).to_csv(
                '{}/bootstrap_matrix.csv'.format(output), sep=",", index=True,
                header=False)

            verboseprint("Save intermediate variance components",
                         verbose=self.verbose)
            verboseprint(("Write covariance matrices based on average of "
                          "bootstrap matrices"), verbose=self.verbose)
            pd.DataFrame(resultsCombineBootstrap['Cg_average']).to_csv(
                "%s/Cg_average_seed%s.csv" % (output, self.seed),
                sep=",", header=False, index=False)
            pd.DataFrame(resultsCombineBootstrap['Cn_average']).to_csv(
                "%s/Cn_average_seed%s.csv" % (output, self.seed),
                sep=",", header=False, index=False)

            verboseprint(("Pickle array of all [`S` x `S`] bootstrap "
                          "covariance matrices"), verbose=self.verbose)
            cPickle.dump(resultsCombineBootstrap['Cg_all_bs'],
                         open('{}/Cg_all_bootstraps.p'.format(output), "wb"))
            cPickle.dump(resultsCombineBootstrap['Cn_all_bs'],
                         open('{}/Cn_all_bootstraps.p'.format(output), "wb"))

            verboseprint(("Pickle result parameters of BFGS fit for fitting "
                          "the [`S` x `S`] bootstrap covariance matrices to "
                          "the [`P` x `P`] overall trait covariance matrices"),
                         verbose=self.verbose)
            cPickle.dump(resultsCombineBootstrap['results_fit_Cg'],
                         open("%s/optimise_results_Cg.p" % (output), "wb"))
            cPickle.dump(resultsCombineBootstrap['results_fit_Cg'],
                         open("%s/optimise_results_Cn.p" % (output), "wb"))

        if self.timing:
            verboseprint("Save process times", verbose=self.verbose)
            pd.DataFrame(resultsCombineBootstrap['proc_time_ind_bs']).to_csv(
                "%s/process_time_all_bootstraps.csv" % (output),
                sep=",",
                header=False,
                index=False)
            overall_time = pd.DataFrame(
                [resultsCombineBootstrap['proc_time_combine_bs'],
                 resultsCombineBootstrap['proc_time_sum_ind_bs']],
                index=["Proctime combine BS",
                       "Proctime sum of individual BS"])
            overall_time.to_csv("%s/process_time_summary.csv" % output,
                                sep=",", header=False, index=True)

    def __generateBootstrapMatrix(self, seed=12321, minCooccurrence=3,
                                  n=None):
        r"""
        Generate subsampling matrix.

        Arguments:
            seed (int, optional):
                for pseudo-random numbers generation; default: 12321
            minCooccurrence (int):
                minimum number a trait pair should be sampled; once reached
                for all trait pairs, sampling is stopped if n is None;
                default=3
            n (int):
                if not None, sets the total number of permutations,
                otherwise n determined by minCooccurrence;  default: None

        Returns:
            None:
                updates LiMMBo instance with:

             - **seed** (int):
               seed for pseudo-random numbers generation
             - **runs** (int):
                n, if n was not None, or determined once all trait-trait
                subsamplings have occurrd minCooccurence times
             - **counts_min** (int):
               minimum trait-trait co-occurrence in sampling matrix
             - **bootstrap_matrix** (numpy.array):
                [`runs` x `S`] matrix containing bootstrap samples of numbers
                range(`P`)
        """
        rand_state = np.random.RandomState(seed)
        counts = sp.zeros((self.P, self.P))
        return_list = []

        if n is not None:
            verboseprint(
                ('Generate bootstrap matrix with {} bootstrap samples '
                 '(number of specified bootstraps').format(n),
                verbose=self.verbose)
            for i in xrange(n):
                bootstrap = rand_state.choice(a=range(self.P), size=self.S,
                                              replace=False)
                return_list.append(bootstrap)
                index = np.ix_(np.array(bootstrap), np.array(bootstrap))
                counts[index] += 1
            self.runs = n
        else:
            while counts.min() < minCooccurrence:
                bootstrap = rand_state.choice(a=range(self.P), size=self.S,
                                              replace=False)
                return_list.append(bootstrap)
                index = np.ix_(np.array(bootstrap), np.array(bootstrap))
                counts[index] += 1
            self.runs = len(return_list)
            verboseprint(
                ('Generated bootstrap matrix with {} bootstrap runs '
                 'such that each trait-trait combination was '
                 'sampled {}').format(self.runs, minCooccurrence),
                verbose=self.verbose)

        self.seed = seed
        self.bootstrap_matrix = np.array(return_list)
        self.counts_min = int(counts.min())

    def __bootstrapPhenotypes(self, bs):
        r"""
        Subsample [`S`] phenotypes with [`N`] samples form total of [`P`]
        phenotypes. Indices for subsampling provided in [`bs` x `S`]
        LiMMBo.bootstrap_matrix, where `bs` is the total number of bootstraps

        Arguments:
            bs (int):
                bootstrap index

        Uses:
            self.bootstrap_matrix (array-like):
                [`bs` x `S`] with subsampling indeces for phenotypes

        Returns:
            (numpy.array):

                - **phenotypes**:
                  [`N` x `S`] of subsampled phenotypes
        """
        bootstrap = self.bootstrap_matrix[bs, :]
        phenotypes = self.phenotypes[:, bootstrap]

        return phenotypes

    def __getBootstrapResults(self, results):
        r"""
        Collect bootstrap results of [`S` x `S`] traits and combine all runs to
        total [`P` x `P`] covariance matrix

        Arguments:
            results (list):
                results of runBootstrapVarianceDecomposition()

        Uses:
            runs (int):
                number of bootstrapping runs executed for this experiment

        Returns:
            (dictionary):
                dictionary containing:

                - **Cg_fit** (numpy.array):
                  [`P` x `P`] genetic covariance matrix via fitting
                - **Cn_fit** (numpy.array):
                  [`P` x `P`] noise covariance matrix via fitting
                - **Cg_average** (numpy.array):
                  [`P` x `P`] genetic covariance matrix via simple average
                - **Cn_average** (numpy.array):
                  [`P` x `P`] noise covariance matrix via simple average
                - **Cg_bs** (numpy.array):
                  [`runs` x `S` x `S`] genetic covariance matrices of `runs`
                  phenotype subsets of size `S`
                - **Cn_bs** (numpy.array):
                  [`runs` x `S` x `S`] noise covariance matrices of `runs`
                  phenotype subsets of size `S`
                - **process_time_bs** (list):
                  run times for all variance decomposition runs of [`S` x `S`]
                  Cg and Cn
                - **number_of_successful_bs** (int):
                  total number of successful bootstrapping runs i.e. variance
                  decomposition converged
                - **results_fit_Cg** ():
                  results parameters of the bfgs-fit of the genetic covariance
                  matrices (via scipy.optimize.fmin_l_bfgs_g)
                - **results_fit_Cn** ():
                  results parameters of the bfgs-fit of the noise covariance
                  matrices (via scipy.optimize.fmin_l_bfgs_g)
        """

        # list to contain trait indices of each bootstrap run
        bootstrap = {}

        # create np.arrays of dimension PxP for mean and std of each entry in
        # covariance matrix
        Cg_average = sp.zeros((self.P, self.P))
        Cn_average = sp.zeros((self.P, self.P))

        # Process time per bootstrap
        process_time_bs = []

        # create np.arrays of dimension runs x P x P to store the bootstrapping
        # results for averaging
        Cg_bs_large = nans((self.runs, self.P, self.P))
        Cn_bs_large = nans((self.runs, self.P, self.P))

        Cg_bs = nans((self.runs, self.S, self.S))
        Cn_bs = nans((self.runs, self.S, self.S))

        n = 0
        for vdresult in results:
            bootstrap[n] = vdresult['bootstrap']
            process_time_bs.append(vdresult['process_time'])

            # store results of each bootstrap as matrix of inflated
            # matrices: NAs for traits that were not sampled
            if vdresult['Cg'] is None or vdresult['Cn'] is None:
                continue
            Cg_bs_large[n, :, :] = inflate_matrix(
                vdresult['Cg'], bootstrap[n], P=self.P, zeros=False)
            Cn_bs_large[n, :, :] = inflate_matrix(
                vdresult['Cn'], bootstrap[n], P=self.P, zeros=False)

            # store results of each bootstrap as matrix of small S x S
            # matrices generated by each bootstrap
            Cg_bs[n, :, :] = vdresult['Cg']
            Cn_bs[n, :, :] = vdresult['Cn']
            n += 1
        # Total number of successful bootstrapping runs
        number_of_bs = n - 1

        # Computing mean of bootstrapping results at
        # each position p1, p2 in overall PxP covariance matrix
        verboseprint(
            ("Computing mean of bootstrapping results"),
            verbose=self.verbose)
        for p1 in range(self.P):
            for p2 in range(self.P):
                vals = Cg_bs_large[:, p1, p2]
                Cg_average[p1, p2] = vals[~sp.isnan(vals)].mean()
                vals = Cn_bs_large[:, p1, p2]
                Cn_average[p1, p2] = vals[~sp.isnan(vals)].mean()

        Cg_reg, ev_g = regularize(Cg_average)
        Cn_reg, ev_n,  = regularize(Cn_average)

        verboseprint(
            ("Fitting bootstrapping results: minimize residual sum of "
             "squares over all bootstraps"),
            verbose=self.verbose)
        Cg_fit, results_fit_Cg = self.__fit_bootstrap_results(
            cov_init=Cg_reg,
            cov_bootstrap=Cg_bs,
            bootstrap_indeces=bootstrap,
            number_of_bs=number_of_bs,
            name="Cg")
        Cn_fit, results_fit_Cn = self.__fit_bootstrap_results(
            cov_init=Cn_reg,
            cov_bootstrap=Cn_bs,
            bootstrap_indeces=bootstrap,
            number_of_bs=number_of_bs,
            name="Cn")

        results = {'Cg_bs': Cg_bs, 'Cn_bs': Cn_bs,
                   'Cg_fit': Cg_fit, 'Cn_fit': Cn_fit,
                   'Cg_average': Cg_average, 'Cn_average': Cn_average,
                   'process_time_bs': process_time_bs,
                   'number_of_successful_bs': number_of_bs,
                   'results_fit_Cg': results_fit_Cg,
                   'results_fit_Cn': results_fit_Cn,
                   }

        return results

    def __fit_bootstrap_results(self, cov_init, cov_bootstrap,
                                bootstrap_indeces, number_of_bs, name):
        r"""
        Fit  bootstrap results of [`S` x `S`] traits and combine all runs to
        total [`P` x `P`] covariance matrix.

        Arguments:
            cov_init (array-like):
                [`P` x `P`] covariance matrix used to initialise fitting
            number_of_bs (int):
                number of successful bootstrapping runs executed for this
                experiment
            cov_bootstrap (array-like):
                [`number_of_bs` x `S` x `S`] bootstrap results
            bootstrap_indeces (array-like):
                [`number_of_bs` x `S`] matrix of bootstrapping indeces
            name (string):
                name of covariance matrix

        Returns:
            (tuple):
                tuple containing:

                - **C_opt_value** (array-like):
                  [`P` x `P`] covariance matrix if fit successful, else 1x1
                  matrix containing string 'did not converge'
                - **results_fit** ():
                  results parameters of the bfgs-fit of the covariance
                  matrix (via scipy.optimize.fmin_l_bfgs_g)


        """

        # initialise free-from Covariance Matrix: use simple average of
        # bootstraps as initial values
        # make use of Choleski decomposition and get parameters associated with
        # cov_init
        C_init = FreeFormCov(self.P)
        C_init.setCovariance(cov_init)
        params = C_init.getParams()

        # initialise free-form covariance matrix C_fit: parameters to be set in
        # optimize function
        C_fit = FreeFormCov(self.P)

        # Fit bootstrap results to obtain closest covariance matrix
        # use parameters obtained from mean-initiated covariance matrix above
        verboseprint(
            "Fitting parameters (minimizing rss via BFGS)...",
            verbose=self.verbose)
        results_fit = opt(
            self.__rss,
            x0=params,
            args=(C_fit, number_of_bs, bootstrap_indeces, cov_bootstrap),
            factr=1e12,
            iprint=2)

        if results_fit[2]['warnflag'] == 0:
            C_opt = FreeFormCov(self.P)
            C_opt.setParams(results_fit[0])
            C_opt_value = C_opt.K()
        else:
            C_opt_value = np.array(['did not converge'])

        return C_opt_value, results_fit

    def __rss(self, params, C, n_bootstrap, index_bootstrap, list_C):
        r"""
        Compute residual sum of squares and gradient for estimate of covariance
        matrix C and bootstrapped values of C.

        Arguments:
            params (array-like):
                of parameters used for initialising estimate C
                (length: 1/2*`P`*(`P`+1))
            C:
                intialised free-form covariance matrix, to be fitted
            n_bootstrap (int):
                number of successful bootstrapping runs executed for this
                experiment
            index_bootstrap (numpy.array):
                [`number_of_bs` x `S`] matrix of bootstrapping indeces
            list_C (list):
                list of n_bootstrap [`S` x `S`] bootstrapped covariance
                matrices

        Returns:
            (tuple):
                tuple containing:

                - **RSS_res**:
                  residual sum of squares
                - **RSS_grad_res**:
                  gradient of residual sum of squares function

        """

        # initialise values of covariance matrix with parameters (based on
        # cholesky decomposition)
        params = np.array(params)
        C.setParams(params)
        # get values of the covariance matrix
        C_value = C.K()
        # number of parameters: 1/2*P*(P+1)
        n_params = params.shape[0]
        # compute residual sum of squares between estimate of P x P Cg and
        # bootstrapped S x S Cgs
        RSS_res, index = self.__RSS_compute(n_bootstrap, C_value, list_C,
                                            index_bootstrap)
        # compute gradient (first derivative of rss at each bootstrap)
        RSS_grad_res = self.__RSS_grad_compute(n_bootstrap, n_params, C,
                                               C_value, list_C, index)

        return RSS_res, RSS_grad_res

    @staticmethod
    def __RSS_compute(n_bootstrap, C_value, list_C, index_bootstrap):
        r"""
        Compute residual sum of squares (rss) for each bootstrap run

        - used parameter to be optimized in quasi-Newton method (BFGS) of
          optimization
        - bootstrap matrices versus matrix to be optimized
        - matrix to be optimized initialised with average over all bootstraps
          for each position

        Arguments:
            n_bootstrap (int):
                number of successful bootstrap runs;
            C_value (array-like):
                [`P` x `P`] matrix to be optimized
            list_C (array-like):
                [`n_bootstrap` x `S` x `S`]  bootstrapped covariance matrices

        Returns:
            (tuple):
                tuple containing:

                - **res** (double):
                  sum of residual sum of squares over all `n_bootstrap` runs
                - **index** (list) :
                  [`n_bootstrap` x `S`] list containing trait indeces used in
                  each bootstrap run; to be passed to gradient computation

        """

        res = 0
        index = {}
        for i in range(n_bootstrap):
            index[i] = np.ix_(index_bootstrap[i], index_bootstrap[i])
            res += bn.ss(C_value[index[i]] - list_C[i], axis=None)
        return res, index

    @staticmethod
    def __RSS_grad_compute(n_bootstrap, n_params, C, C_value, list_C, index):
        r"""
        Compute gradient of residual sum of squares (rss) for each bootstrap
        run at each index

        - used as gradient parameter in quasi-Newton method (BFGS) of
          optimization
        - matrix to be optimized initialised with average over all bootstraps
          for each position

        Arguments:
            n_bootstrap (int):
                number of successful bootstrap runs;
            n_params (int):
                number of parameters of the model (parameters needed to build
                positive, semi-definite matrix with Choleski decomposition)
            C_value (array-like):
                [`P` x `P`] matrix to be optimized
            list_C (array-like):
                [`n_bootstrap` x `S` x `S`]  bootstrapped covariance matrices
            index (list):
                [`n_bootstrap `x `S`] trait indices used in each bootstrap run

        Returns:
            (numpy.array):

                - **res**:
                  [`n_params` x 1] sum of gradient over all successful
                  bootstrap runs for each parameter to be fitted.

        """
        res = sp.zeros(n_params)
        for pi in range(n_params):
            Cgrad = C.K_grad_i(pi)
            for i in range(n_bootstrap):
                res[pi] += (2 * Cgrad[index[i]] *
                            (C_value[index[i]] - list_C[i])).sum()
        return res

    def __VarianceDecomposition(self, phenoSubset, bs=None):
        r"""
        Compute variance decomposition of phenotypes into genetic and noise
        covariance

        Arguments:
            phenoSubset (array-like):
                [`N` x `S`] phenotypes for which variance decomposition should
                be computed
            bs (int):
                number of subsample
            phenotypes (array-like):
                [`N` x `P`] original phenotypes
            relatedness (array-like):
                [`N` x `N`] kinship/genetic relatedness used in estimation of
                genetic component
            output (string):
                output directory; needed for caching

        Returns:
            (dictionary):
                dictionary containing:

                - **Cg** (numpy.array):
                  [`S` x `S`] genetic variance component
                - **Cn** (numpy.array):
                  [`S` x `S`] noise variance component
                - **process_time** (double):
                  cpu time of variance decomposition
                - **bsindex** (int):
                  number of subsample
        """

        # time variance decomposition
        t0 = time.clock()
        vd = limix.mtset.MTSet(Y=phenoSubset, R=self.relatedness)
        vd_null_info = vd.fitNull(
            cache=False,
            n_times=self.iterations,
            rewrite=True)
        t1 = time.clock()
        processtime = t1 - t0

        if vd_null_info['conv']:
            verboseprint(
                ('Variance decomposition for bootstrap number {} '
                 'converged').format(bs + 1), verbose=self.verbose)
            Cg = vd_null_info['Cg']
            Cn = vd_null_info['Cn']
        else:
            verboseprint(
                ('Variance decomposition for bootstrap number {} '
                 'did not converge').format(bs + 1), verbose=self.verbose)
            Cg = None
            Cn = None
        return {'Cg': Cg, 'Cn': Cn, 'process_time': processtime, 'bsindex': bs}



from setuptools import setup

if __name__ == '__main__':
    readme = open('README.md').read()
    import pypandoc
    long_description = pypandoc.convert_text(
        readme, 'rst', format='markdown_github')

    setup(long_description=long_description,
          entry_points = {
              'console_scripts': [
                  'runAssociation=limmbo.bin.runAssociation:entry_point',
                  'runVarianceEstimation=limmbo.bin.runVarianceEstimation:entry_point'
                  ]
              }
          )
